/*---------------------------------------------------------------------------*/
/* ATAPI AUDIO PLAYER using ATAPI DRIVER for MINIX --------------------------*/
/* Written by LEDOUX Alexis -- 1999 -----------------------------------------*/
/*****************************************************************************/

/*----------------------------------------------------------------------------*/
/* Header files --------------------------------------------------------------*/
/******************************************************************************/
#include <sys/types.h>
#include <signal.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <curses.h>
#include <minix/cdrom.h>
#include <minix/atapi.h>

/*----------------------------------------------------------------------------*/
/* Global variables ----------------------------------------------------------*/
/******************************************************************************/
char *string;			/* Path for the playlist file */ 
int  oldpos,pos=0,X=1;		/* Move coordinates */ 
int  fd;                        /* Handle for the Device */
int  startlist=1;               /* Actual playing track */
int  trackf;                    /* Temporary variables */
int  track_list[55];            /* Array to store the playing order */
int  state=CDSTOP;              /* State of the device */  
int  intro=NO,random=NO; 
int  current_cd=0;
int  nb_cd=0;

struct cd playlist[CDSIZE];	/* Playlists of the user */ 
struct atapitoc Toc;            /* Table of contents */
struct cd_play_track playtrack; /* Structure to play a track */
struct cd_play_mss playmss;	/* Structure to play from mss_start to mss_end */
struct atapi_subchnl time;      /* Structure of the subchannel */

/* Manage the button text */
char *position[11]={" |<"," <<", " " ," >" ," >>"," >|", " x" ," x"};

/*----------------------------------------------------------------------------*/
/* Prototype of every functions ----------------------------------------------*/
/******************************************************************************/
_PROTOTYPE(void Initialise, (void));        /* Initialise terminal */
_PROTOTYPE(int  Abort,      (char *s));     /* Exit program and reset term */
_PROTOTYPE(void PrintMsg,   (char *Str));   /* Print a message in high white */
_PROTOTYPE(void ClearMsg,   (void));        /* Clear previous message */
_PROTOTYPE(void WriteCommand,(char *S));    /* Highlight first letter of S */
_PROTOTYPE(void PlayBackground,(void));     /* Play audio CD in background */
_PROTOTYPE(char GetChar,    (void));        /* Get character until valid */
_PROTOTYPE(void Getstring, (int n,char *target));	    /* Get the title of the track */
_PROTOTYPE(char GetAChar,   (void));	    /* Get a character until valid */
_PROTOTYPE(void ReadCommand,(char *Command)); /* Read command for the character */
_PROTOTYPE(void InterpretCommand,(int Command)); /* Interpret command */
_PROTOTYPE(void fill_toc,   (void) );   /* Display Table of content in window */ 
_PROTOTYPE(void get_playlist,(void) );     /* Get playlists */
_PROTOTYPE(void save_playlist,(void) );     /* Save playlists */
_PROTOTYPE(int  execute,    (int) );        /* execute ATAPI Commands */
_PROTOTYPE(void main,       (int argc,char **argv)); /* Main function */

/*----------------------------------------------------------------------------*/
/* Initialisation function ---------------------------------------------------*/
/******************************************************************************/ 
void Initialise(void)
{
  initscr(); 
  raw();
  noecho();
}

/*----------------------------------------------------------------------------*/
/* Reset terminal and exit from the program. ---------------------------------*/
/******************************************************************************/
int Abort(char *s)
{ 
   printf("\x1b[2J\x1b[H\x1b[m");
   move(LINES - 1,0);
   refresh();
   endwin();
   exit(0);
}

/*----------------------------------------------------------------------------*/
/* Clear and Prints a message ------------------------------------------------*/
/******************************************************************************/
void PrintMsg(char *Str)
{ 
   printf("\x1b[9;5H\x1b[1;37;40m\x1b[2K");
   printf("\x1b[9;5H\x1b[1;37;40m%s",Str);
}

/*----------------------------------------------------------------------------*/
/* Get the title of the track ------------------------------------------------*/
/******************************************************************************/
void Getstring(int n,char *target)
{
   int i,tip=0;
   char c,string[1024];
   for(i=0;i<n;i++)
   {
      c=GetAChar();
      if(c=='\n') break;
      else
         if(c!='?')
            if(c=='#')
            {
               printf("%c %c",8,8);
               c='\0';
               string[tip]='\0';
               tip--;
            }else{
               printf("%c",c);
               string[tip]=c;
               tip++;
            }
   } 
   string[tip]='\0';
   strcpy(target,string);
}

/*----------------------------------------------------------------------------*/
/* Highlights the first letter of the string S -------------------------------*/
/******************************************************************************/
void WriteCommand(char *S)
{
  printf("\x1b[1;36;40m%c",S[0]);
  printf("\x1b[0;36;40m%s",S+1);
}

/*----------------------------------------------------------------------------*/
/* Draw the spash screen and wait for a key press ----------------------------*/
/******************************************************************************/
void Draw_Splash(void)
{
  printf("\x1b[2J");

  printf( "\x1b[7;22H\x1b[0;30;46mͻ");
  printf( "\x1b[8;22H\x1b[0;30;46m   \x1b[1;36;46m A T A P I   C D - P L A Y E R \x1b[0;30;46m   "); 
  printf( "\x1b[9;22H\x1b[0;30;46m                                     ");
  printf("\x1b[10;22H\x1b[0;30;46m      Written  by Alexis LEDOUX      ");
  printf("\x1b[11;22H\x1b[0;30;46m                                     ");
  printf("\x1b[12;22H\x1b[0;30;46m           IT Carlow  1999           "); 
  printf("\x1b[13;22H\x1b[0;30;46m             Version 1.0             ");
  printf("\x1b[14;22H\x1b[0;30;46m                                     ");
  printf("\x1b[15;22H\x1b[0;30;46m      \x1b[1;33;46m<Press a key to continue>\x1b[0;30;46m      ");
  printf("\x1b[16;22H\x1b[0;30;46mͼ\x1b[0;37;40m");
}

/*----------------------------------------------------------------------------*/
/* Draw the user interface ---------------------------------------------------*/
/******************************************************************************/
void Draw_interface(void)
{
  int i; 
  printf("\x1b[2J");

  printf("\x1b[1;2H\x1b[0;30;46mͻ");
  printf("\x1b[2;2H\x1b[0;30;46m   "); 
  printf("\x1b[1;36;46m A T A P I  C D - P L A Y E R ");
  printf("\x1b[0;30;46m   ");
  printf("\x1b[0m     \x1b[1;36;40mW\x1b[1;37;40mritten  by \x1b[1;36;40mA\x1b[1;37;40mlexis LEDOUX");
  printf("\x1b[3;2H\x1b[0;30;46m                                    ");
  printf("\x1b[0m     \x1b[1;37;40m1999    4th year Computing");
  printf("\x1b[4;2H\x1b[0;30;46m  Track : \x1b[1;36;46m00     00:00\x1b[0;30;46m   Random  x  ");
  printf("\x1b[5;2H\x1b[30;46m                         Intro   x  ");
  printf("\x1b[6;2H\x1b[30;46m                                    ");
  printf("\x1b[7;2H\x1b[30;46m     |<   <<      >   >>   >|      ");
  printf("\x1b[8;2H\x1b[30;46mͼ\x1b[0;37;40m");

  printf("\x1b[10;2H   \x1b[30;46mͻ\x1b[0;37;40m");
  printf("\x1b[11;2H   \x1b[30;46m                                                                    \x1b[0;37;40m");
  printf("\x1b[12;2H   \x1b[30;46m N  Title                                                    Time  \x1b[0;37;40m");
  printf("\x1b[13;2H   \x1b[30;46m͹\x1b[0;37;40m");
  printf("\x1b[14;2H");
  for(i=0;i<10;i++) 
     printf("   \x1b[30;46m                                                                    \x1b[0;30;40m\x1b[1B\x1b[75D"); 
  printf("\x1b[24;2H   \x1b[30;46mͼ\x1b[0;37;40m");

  printf("\x1b[5;45H");        WriteCommand("+/");
  printf("\x1b[5;47H");        WriteCommand("- Scroll title page");
  printf("\x1b[6;45H");        WriteCommand("Eject");
  printf("\x1b[6;62HShow ");   WriteCommand("PlayList");
  printf("\x1b[7;45H");        WriteCommand("Save PlayList");
  printf("\x1b[7;62HRun in "); WriteCommand("Background");
  printf("\x1b[8;45H");        WriteCommand("Quit"); 
}

/*----------------------------------------------------------------------------*/
/* Change the order of the playlist ------------------------------------------*/
/******************************************************************************/
void play_rand(void)
{
   int next=1;
   int i,j;

   while(next < Toc.Hdr.LastTrk)
   {
      track_list[next] = 1;
      i=(rand() % (Toc.Hdr.LastTrk - Toc.Hdr.FirstTrk + 1 )+1);
      for (j=0;j<next;j++)
        if(track_list[j] == i) break;
      if (track_list[j] !=i) track_list[next++]=i;
   }
   for(j=0;j<Toc.Hdr.LastTrk;j++) track_list[j]-=1;
   random=YES;
}

/*----------------------------------------------------------------------------*/
/* Play the CD in normal or random mode in Background ------------------------*/
/******************************************************************************/
void PlayBackground(void)
{
   int sleeptime,i;
   
   switch(fork())
   {
      case -1: PrintMsg("Fork not possible");
               Abort("Bye Bye");
      case 0: break;  /* child continue */
      default : Abort("Bye Bye");   /* Parent stop */
   }

   if(random==NO) 
      if(intro==YES)
      {
         for(i=Toc.Hdr.FirstTrk;i<=Toc.Hdr.LastTrk;i++)
         {
            playtrack.begin_track=track_list[i];
            playtrack.end_track=track_list[i];
            ioctl(fd,CDIOPLAYTI,&playtrack);
            sleep(20); 
         }
      }else{ 
         ioctl(fd,CDIOREADTOC,&Toc);
         if (state!=CDPLAY)
         { 
            playtrack.begin_track=Toc.Hdr.FirstTrk;
            playtrack.end_track=Toc.Hdr.LastTrk;
            ioctl(fd,CDIOPLAYTI,&playtrack);
         }
         printf("Thanks for using ACDPlayer\nIt is playing in the Background\n<Press any key to Continue>"); 
         exit(0);
      }  
   else
      for(i=Toc.Hdr.FirstTrk;i<=Toc.Hdr.LastTrk;i++)
      {
         playtrack.end_track=playtrack.begin_track=track_list[i]; 
         ioctl(fd,CDIOPLAYTI,&playtrack);
         if (intro==YES) sleep(20);
         else
         {
            sleeptime=(Toc.Ent[track_list[i]].Where.Time.Min*60+Toc.Ent[track_list[i]].Where.Time.Sec)
               -(Toc.Ent[track_list[i]-1].Where.Time.Min*60+Toc.Ent[track_list[i]-1].Where.Time.Sec);
            sleeptime--;
            sleep(sleeptime);
            while(ioctl(fd,CDIOREADTIME,&time)==1);
         }
      }   
   ioctl(fd,CDIOSTOP,0);
   exit(0);
}

/*----------------------------------------------------------------------------*/
/* Get a character from the keyboard -----------------------------------------*/
/******************************************************************************/
char GetAChar(void)
{
  int c; 
  c = getch(); 
  if (c < 0) abort();
  if (c == '\033') {  /* ARROW Key */
     if ((c = getch()) == '[') {
	c = getch();
	switch (c) {
	   case 'A': c = '?'; break;  /* UP    */
	   case 'B': c = '?'; break;  /* DOWN  */
	   case 'C': c = '?'; break;  /* RIGHT */
	   case 'D': c = '?'; break;  /* LEFT  */
	   default:  c = '?'; break;
	}
     }else c = '?';
  }
  if(c==13) c='\n';
  if(c==8) c='#'; 
  return c;
}

/*----------------------------------------------------------------------------*/
/* Get a character from the keyboard -----------------------------------------*/
/******************************************************************************/
char GetChar(void)
{
  int c;

  c = getch(); 
  if (c < 0) abort();
  if (c == '\033')
  {  /* ARROW Key */
     if ((c = getch()) == '[')
     {
	c = getch();
	switch (c) {
	   case 'A': c = 'U'; break;  /* UP    */
	   case 'B': c = 'D'; break;  /* DOWN  */
	   case 'C': c = 'R'; break;  /* RIGHT */
	   case 'D': c = 'L'; break;  /* LEFT  */
	   default:  c = '?'; break;
	}
     }else c = '?';
  } 
  if (islower(c)) return toupper(c);
  else return c;
}

/*----------------------------------------------------------------------------*/
/* Print the button when active or inactive ----------------------------------*/
/******************************************************************************/
void PrintPos(int post,int st)
{
   int show=0;

   if (st==NEW) show=1;
   switch(post)
   {
      case 0:  if(show) PrintMsg("Previous Track");printf("\x1b[7;7H");break;
      case 1:  if(show) PrintMsg("5 secondes less");printf("\x1b[7;12H");break;
      case 2:  if(show) PrintMsg("Stop Track");printf("\x1b[7;17H");break;
      case 3:  if((show)&&(state==CDPLAY)) PrintMsg("Pause");
               if((show)&&(state==CDPAUSE)) PrintMsg("Resume");
               if((show)&&(state==CDSTOP)) PrintMsg("Play Track");
               printf("\x1b[7;21H");break;
      case 4:  if(show) PrintMsg("5 secondes later");printf("\x1b[7;25H");break;
      case 5:  if(show) PrintMsg("Next Track");printf("\x1b[7;30H");break;
      case 6:  if(show) PrintMsg("Random (only available in Background)");printf("\x1b[4;35H");break;
      case 7:  if(show) PrintMsg("Intro (only available in Background)");printf("\x1b[5;35H");break;
      case 8:  if(show) PrintMsg("Play actual Track");printf("\x1b[14;7H");break;
      case 9:  if(show) PrintMsg("Edit Title of Track");printf("\x1b[14;12H");break;
      case 10: if(show) PrintMsg("Play actual Track");printf("\x1b[14;70H");break;
      default: return;
   }
   if ((st==OLD) && (post<8))
      printf("\x1b[0;30;46m%s\x1b[0;30;46m",position[post]);
   if ((st==NEW) && (post<8)) 
      printf("\x1b[1;36;46m%s\x1b[0;30;46m",position[post]);  
}

/*----------------------------------------------------------------------------*/
/* Reads in a valid command character ----------------------------------------*/
/******************************************************************************/
void ReadCommand(char *Command)
{
  int ValidCommand;

  do{
	ValidCommand = TRUE;

	PrintPos(oldpos,OLD);
        PrintPos(pos,NEW);
        
        *Command = GetChar();	/* Read from keyboard */
	switch (*Command)
	{
	    /* Key Action */
	    case '\n':		/* '\n' or space */
	    case ' ' :
               break;

            /* Commands Key */
	    case 'E': PrintMsg("Eject/Close the Door");break;
	    case 'S': PrintMsg("Save PlayList");break;
	    case 'P': PrintMsg("Playlist");break;
	    case 'B': PlayBackground();break;
	    case 'Q': PrintMsg("Bye Bye !!");break; 

	    /* POSITION */
	    case 'L':
	    case 'R':
	    case 'U':
	    case 'D': break;
	    case '8': *Command = 'U'; break;
	    case '2': *Command = 'D'; break;
	    case '4': *Command = 'L'; break;
	    case '6': *Command = 'R'; break;

	    case '-': 
	    case '+': break;

	    default:break;
       }
   }while (!ValidCommand);
}
/*----------------------------------------------------------------------------*/
/* Interpret the command -----------------------------------------------------*/
/******************************************************************************/
void InterpretCommand(Command)
char Command;
{
  int i;

  switch (Command) {

      /* Move playlist down */
      case '-': startlist>1?startlist--:startlist;
                fill_toc();break;

      /* Move playlist Up */
      case '+': startlist+10<=Toc.Hdr.LastTrk?startlist++:startlist;
                fill_toc();break;

      /* LEFT */
      case 'L':
        oldpos=pos;
	if (pos>=1) { pos--; if (pos<8) X=1;
	}else{ pos=8; X=1; }
        break;

      /* RIGHT */
      case 'R':
        oldpos=pos;
	if (pos<10) { pos++; if (pos<8) X=1; }
	else{ pos=0; X=1; }
	break;

      /* DOWN */
      case 'D':
        oldpos=pos;
        if ((pos>7 && pos<=10) && X<Toc.Hdr.LastTrk) X = X + 1;
        startlist=X;
        fill_toc(); 
        break;

      /* UP */
      case 'U':
        oldpos=pos;
        if ((pos>7 && pos<=10) && X>1) X = X - 1;
        startlist=X;
        fill_toc(); 
        break;

      /* SPACE PRESSED */
      case ' ':

        /* RANDOM & INTRO */
        if (pos==6||pos==7)
        {
          if (strcmp(position[pos]," ")==0)
          {
             position[pos]=" x";
             if (pos==6)
             {
                for(i=0;i<Toc.Hdr.LastTrk;i++) track_list[i]=i;
                random=NO;
             }else intro=NO;
          }
          else
          {
             position[pos]=" ";
             if (pos==6) play_rand();
             else intro=YES;
          }
          fill_toc();
          break;
        }

        /* 5 seconds more or less in the song */ 
        if ((pos==1)||(pos==4))
        {
           if (state==CDPLAY)
           {
              execute(CDIOREADTIME);
              if (pos==1){
                 if (time.absaddr.msf.Sec>=5) time.absaddr.msf.Sec-=5;
                 else
                 {
                    time.absaddr.msf.Min>0?time.absaddr.msf.Min-=1:0;
                    time.absaddr.msf.Sec+=55;
                 }
              }
              if (pos==4){
                 if (time.absaddr.msf.Sec<55) time.absaddr.msf.Sec+=5;
                 else
                 {
                    time.absaddr.msf.Min+=1;
                    time.absaddr.msf.Sec-=55; 
                 }
              }
              playmss.begin_mss[0]=time.absaddr.msf.Min;
              playmss.begin_mss[1]=time.absaddr.msf.Sec;
              playmss.begin_mss[2]=time.absaddr.msf.Frame;
              playmss.end_mss[0]=Toc.Ent[Toc.Hdr.LastTrk].Where.Time.Min;
              playmss.end_mss[1]=Toc.Ent[Toc.Hdr.LastTrk].Where.Time.Sec;
              playmss.end_mss[2]=Toc.Ent[Toc.Hdr.LastTrk].Where.Time.Frame-1; 
              execute(CDIOPLAYMSS);
              execute(CDIOREADTIME);
              printf("\x1b[4;20H\x1b[1;36;46m%02d:%02d",time.reladdr.msf.Min,time.reladdr.msf.Sec);
              printf("\x1b[4;13H\x1b[1;36;46m%02d\x1b[0;30;46m",time.trk);
              trackf=time.trk; 
           }
           break;
        }

        /* PLAY */
        if (pos==3)
        {
           execute(CDIOREADTOC);
           if (state==CDPLAY)
           {
              execute(CDIOPAUSE); 
              break;
           }
           if (state==CDPAUSE)
           {
              execute(CDIORESUME); 
              break;
           }
           if (state==CDSTOP)
           {
              execute(CDIOSTOP); 
              trackf=playtrack.begin_track=Toc.Hdr.FirstTrk;
              playtrack.end_track=Toc.Hdr.LastTrk; 
              execute(CDIOPLAYTI);
              break;
           }
           if (state==CDNEXT || state==CDPREVIOUS)
           { 
              execute(CDIOPLAYTI);  
              break;
           }
           break;
        }

        /* PREVIOUS TRACK */
        if (pos==0)
        { 
           if (trackf>1) trackf--;
           else trackf=1;
           printf("\x1b[4;13H\x1b[1;36;46m%02d\x1b[0;30;46m",trackf);
           printf("\x1b[4;20H\x1b[1;36;46m00:00");
           playtrack.begin_track=trackf;
           playtrack.end_track=Toc.Hdr.LastTrk; 

           if ((state==CDPLAY)||(state==CDPAUSE)) execute(CDIOPLAYTI);
           else state=CDPREVIOUS; 
           break;
        }

        /* STOP */
        if (pos==2)
        {
           if (state==CDPLAY || state==CDPAUSE) execute(CDIOSTOP);
           break;
        }

        /* NEXT TRACK */
        if (pos==5)
        { 
           if (trackf<Toc.Hdr.LastTrk) trackf++;
           else trackf=Toc.Hdr.LastTrk; 
           printf("\x1b[4;13H\x1b[1;36;46m%02d\x1b[0;30;46m",trackf);
           printf("\x1b[4;20H\x1b[1;36;46m00:00");
           playtrack.begin_track=trackf;
           playtrack.end_track=Toc.Hdr.LastTrk;

           if (state==CDPLAY) execute(CDIOPLAYTI);
           else state=CDNEXT; 
           break;
        }

        /* Play asked track */
        if (pos==8 || pos==10)
        { 
           playtrack.end_track=playtrack.begin_track=track_list[startlist]; 
           printf("\x1b[4;20H\x1b[1;36;46m00:00");
           execute(CDIOPLAYTI); 
           break; 
        }

        /* Get track title */
        if (pos==9)
        {
           printf("\x1b[14;12H\x1b[1;36;46m");
           if(current_cd==CDSIZE)
           {
              current_cd=nb_cd;
              playlist[current_cd].start=Toc.Ent[Toc.Hdr.FirstTrk].Where.Time.Min*100+
                    Toc.Ent[Toc.Hdr.FirstTrk].Where.Time.Sec;
              playlist[current_cd].end =(Toc.Ent[Toc.Hdr.LastTrk].Where.Time.Min*100+
                    Toc.Ent[Toc.Hdr.LastTrk].Where.Time.Sec)-
                   (Toc.Ent[Toc.Hdr.LastTrk-1].Where.Time.Min*100+
                    Toc.Ent[Toc.Hdr.LastTrk-1].Where.Time.Sec);  
              playlist[current_cd].Numtrack=Toc.Hdr.LastTrk-Toc.Hdr.FirstTrk+1;
              nb_cd++;
           }
           memset(playlist[current_cd].track[startlist],' ',55); 
           Getstring(50,playlist[current_cd].track[startlist]);
           /*echo(); scanf("%[ -z]",playlist[current_cd].track[startlist]); noecho();*/
           fill_toc();
           break;
        }

      /* Eject or Close the Door */
      case 'E':
         if (state!=CDEJECT) execute(CDIOEJECT);
         else{
            execute(CDIOCLOSE); 
            sleep(1);
            execute(CDIOREADTOC);
         }
         break;

      /* Read the playlist */
      case 'P':
         execute(CDIOREADTOC);
         break;

      /* Save the actual playlist */
      case 'S':
         PrintMsg("Enter the Title of the CD");
         printf("\x1b[11;12H\x1b[1;36;46m");
         /*echo(); scanf("%[ -z]",&playlist[current_cd].title); noecho();*/
         Getstring(50,playlist[current_cd].title);
         PrintMsg("Saving...");
         save_playlist();
         PrintMsg("Playlist Saved");
         pos=3;
         PrintPos(pos,NEW);
         break;
   }
}

/*---------------------------------------------------------------------------*/
/* Save playlists to the .acdplay file in the home directory ----------------*/
/*****************************************************************************/
void save_playlist(void)
{
   int i=0,j;
   FILE *titles; 
   int finish=FALSE;

   titles=fopen(string,"w");

   while(finish==FALSE)
   {
      fprintf(titles,"[%0004d %0004d %02d]\n",playlist[i].start,playlist[i].end,playlist[i].Numtrack);   
      fprintf(titles,"%40s\n",playlist[i].title); 
      for(j=1;j<=playlist[i].Numtrack;j++) 
         fprintf(titles,"%02d %55s\n",j,playlist[i].track[j]); 
      i++;
      if (playlist[i].start==0&&playlist[i].end==0) finish=TRUE;
   }  
   fclose(titles);
}

/*---------------------------------------------------------------------------*/
/* Read playlists from the .acdplay file in the home directory --------------*/
/*****************************************************************************/
void get_playlist(void)
{
   int i=0,finish=FALSE,temp,j;
   FILE *titles;                   /* Handle to the playlist file */
   int Numtrack;

   titles=fopen(string,"r");  

   while(finish==FALSE)
   {
      if((fscanf(titles,"[%0004d %0004d %02d]\n",&playlist[i].start,&playlist[i].end,&Numtrack))==EOF) 
      {
         finish=TRUE; 
         playlist[i].start=0;
         playlist[i].end=0;
         playlist[i].Numtrack=0;
      }
      else
      { 
         playlist[i].Numtrack=Numtrack;
         fscanf(titles,"%40[ -z]\n",&playlist[i].title); 
         for(j=1;j<=Numtrack;j++)
         {
            for(temp=0;temp<3;temp++) fgetc(titles);
            for(temp=0;temp<55;temp++) playlist[i].track[j][temp]=fgetc(titles); 
            fgetc(titles);
         } 
         i++;
      } 
   } 
   nb_cd=i;
   fclose(titles);
}

/*----------------------------------------------------------------------------*/
/* Display track numbers and their duration ----------------------------------*/
/******************************************************************************/
void fill_toc(void)
{
   int i,time,temp,j;
   int start,end,Numtrack; 
   char c;
   FILE *titles;                   /* Handle to the playlist file */

   /* Get the time of the first and last track of the CD to make the
      a special Header for the playlist. The playlist is a file with
      all the title of your CD's and to recognize them we put first 
      the time of the first and last track in second followed by the
      number of track on the CD */ 

   start=Toc.Ent[Toc.Hdr.FirstTrk].Where.Time.Min*100+
         Toc.Ent[Toc.Hdr.FirstTrk].Where.Time.Sec;
   end =(Toc.Ent[Toc.Hdr.LastTrk].Where.Time.Min*100+
         Toc.Ent[Toc.Hdr.LastTrk].Where.Time.Sec)-
        (Toc.Ent[Toc.Hdr.LastTrk-1].Where.Time.Min*100+
         Toc.Ent[Toc.Hdr.LastTrk-1].Where.Time.Sec); 

   /* Read in the playlist file */
   for(current_cd=0;current_cd<CDSIZE;current_cd++)
      if((start==playlist[current_cd].start)&&(end==playlist[current_cd].end))
         break;

   /* Update the screen */
   printf("\x1b[11;12H\x1b[1;36;46m");
   if (current_cd!=CDSIZE)
      for(i=0;i<40;i++) printf("%c",playlist[current_cd].title[i]); 
   printf("\x1b[14;0H");
   for(i=0;i<10;i++)
   {
      if (startlist+i<=Toc.Hdr.LastTrk)
      { 
         printf("\x1b[6C%02d",track_list[startlist+i]);
         time=(Toc.Ent[track_list[startlist+i]].Where.Time.Min*60+Toc.Ent[track_list[startlist+i]].Where.Time.Sec)
             -(Toc.Ent[track_list[startlist+i]-1].Where.Time.Min*60+Toc.Ent[track_list[startlist+i]-1].Where.Time.Sec); 
         printf("\x1b[3C"); 
         for(j=0;j<55;j++) 
         {
            if ((playlist[current_cd].track[track_list[startlist+i]][j]!='\0')&&(current_cd!=CDSIZE)) 
               printf("%c",playlist[current_cd].track[track_list[startlist+i]][j]);
            else
               printf(" ");
         }
         printf("\x1b[3C%02d:%02d",time/60,time%60);
      }else{
         printf("\x1b[6C\x1b[0;36;46m%-2s","0");
         printf("\x1b[3C\x1b[0;36;46m%-55s","0");
         printf("\x1b[3C\x1b[0;36;46m%-5s","0");
      } 
      printf("\x1b[1B\x1b[75D");
   } 
}

/*---------------------------------------------------------------------------*/
/* Execute the IOCTL Command ------------------------------------------------*/
/*****************************************************************************/
int execute(int command)
{
   int i;
   int result;
   struct atapitoc ChkToc;

   switch(command)
   {
      case CDIOLOCK    : ioctl(fd,CDIOLOCK,0);
                         break;

      case CDIOUNLOCK  : ioctl(fd,CDIOUNLOCK,0);
                         break;

      case CDIOPLAYMSS : ioctl(fd,CDIOSTOP,0);
                         ioctl(fd,CDIOPLAYMSS,&playmss);
                         state=CDPLAY;
                         break;

      case CDIOSTOP    : ioctl(fd,CDIOSTOP,0); 
                         trackf=1; 
                         printf("\x1b[4;13H\x1b[1;36;46m%02d\x1b[0;30;46m",trackf);
                         position[3]=" >";
                         PrintPos(3,OLD);
                         state=CDSTOP;
                         execute(CDIOUNLOCK);
                         break;
                         
      case CDIOEJECT   : ioctl(fd,CDIOSTOP,0);
                         execute(CDIOUNLOCK);
                         ioctl(fd,CDIOEJECT,0);
                         execute(CDIOLOCK);
      			 state=CDEJECT;
                         break;

      case CDIOCLOSE   : execute(CDIOUNLOCK);
      			 ioctl(fd,CDIOCLOSE,0);
                         execute(CDIOLOCK);
                         state=CDSTOP;
                         break;

      case CDIOREADTOC : ChkToc=Toc; 
                         ioctl(fd,CDIOREADTOC,&Toc);
                         if(Toc.Ent[Toc.Hdr.LastTrk].Where.Time.Sec!=ChkToc.Ent[ChkToc.Hdr.LastTrk].Where.Time.Sec)
                         {
                            PrintMsg("Audio CD Change");
                            for(i=Toc.Hdr.FirstTrk;i<=Toc.Hdr.LastTrk;i++) track_list[i]=i; 
                            position[3]=" >";
                            PrintPos(3,OLD);
                            startlist=1;
                            trackf=1;
                            fill_toc(); 
                         }
                         /*execute(CDIOLOCK);*/
                         break;

      case CDIOPLAYTI  : /*ioctl(fd,CDIOSTOP,0);*/
                         /*execute(CDIOREADTOC);*/
                         ioctl(fd,CDIOPLAYTI,&playtrack);
                         printf("\x1b[4;13H\x1b[1;36;46m%02d\x1b[0;30;46m",trackf);
                         position[3]=" ";
                         PrintPos(3,OLD);
                         state=CDPLAY;
                         break;

      case CDIOPAUSE   : ioctl(fd,CDIOPAUSE,0);
                         PrintMsg("Pause");
                         position[3]=" >"; 
                         state=CDPAUSE;
                         break;

      case CDIORESUME  : ioctl(fd,CDIORESUME,0);
                         PrintMsg("Resume");
                         position[3]=" "; 
                         state=CDPLAY;
                         break;

      case CDIOREADTIME: ioctl(fd,CDIOREADTIME,&time);
                         printf("\x1b[4;20H\x1b[1;36;46m%02d:%02d"
                                ,time.reladdr.msf.Min,time.reladdr.msf.Sec);
                         break;
   }
}

/*----------------------------------------------------------------------------*/
/* Main function -------------------------------------------------------------*/
/******************************************************************************/
void main(int argc,char **argv)
{
  char *device; 
  char Command;			/* Command from keyboard */ 

  /* Copy the value of the first argument in the device variable */
  if (argc==2) device=argv[1];
  else device="/dev/cd0";

  /* Initialise display and variables and Draw the interface */
  Initialise(); 
  Draw_Splash();
  getch(); 
  Draw_interface();

  /* Initialise ATAPI CDROM */
  PrintMsg("Finding and Initialising ATAPI CDROM ...");
  if ((fd=open(device,O_RDONLY))<0)
  {
     PrintMsg("No ATAPI CDROM Found"); 
     Abort("Bye Bye");
  }

  PrintMsg("Welcome to ATAPI CD-PLAYER");

  /* Initialise the file to save playlist */
  string=getenv("HOME");
  strcat(string,".acdplay");
  get_playlist();

  /* Read the Toc if it is possible */
  execute(CDIOREADTOC);

  /* Interpret commands until the command 'Quit' */
  do{
     ReadCommand(&Command);
     InterpretCommand(Command);
  } while (Command != 'Q');

  /* Stop the AUDIO CD if playing */
  execute(CDIOSTOP);

  /* Put back the screen to work */
  Abort("Good bye!");
}
