/*---------------------------------------------------------------------------*/
/* ISO9660 Explorer 1999 using ATAPI DRIVER for MINIX -----------------------*/
/* Written by LEDOUX Alexis -- 1999 -----------------------------------------*/
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
/* Header files -------------------------------------------------------------*/
/*****************************************************************************/
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <curses.h>
#include <minix/cdrom.h>
#include <minix/atapi.h>

#define BLOCK_SIZE 2048
#define RED	1
#define WHITE	0

/*---------------------------------------------------------------------------*/
/* Global variables ---------------------------------------------------------*/
/*****************************************************************************/
int oldpos,pos=0;		            /* Move coordinates */ 
int fd;                                     /* Handle for the Device */ 
int High_Sierra = 0;                        /* 1 = high sierra format */
int Iso9660 = 0;                            /* 1 = iso9660 format */
int AudioCd= 0;                             /* 1 = Audio-CD format */ 
int path_length;			    /* Length of the path */
int nr_files=0;                             /* Number of files in the path */
int nr_dir=0;                               /* Number of directory */
int nr_audio=0;                             /* Number of audio tracks */
int Read_Dir=1;                             /* Flag to read directory */ 
int Read_File=0;                            /* Flag to read a file */
int Aflag=1;                                /* Flag to save text file UNIX format */
int startlist = 0;                          /* Variable for the interface */
int totsize=0;				    /* Total size in bytes selected */
int nb=0;				    /* Nb of files selected */

char Buffer[BLOCK_SIZE];	            /* buffer to hold read data */  
char path[80];		                    /* Path to lookup */
char target[80];               		    /* String to save the target path */

struct iso9660_descriptor *Iso_Vol_Desc;    /* iso9660 volume descriptor */ 
struct high_sierra_descriptor *Hs_Vol_Desc; /* high sierra volume descriptor */ 
struct atapitoc Toc;                        /* Table of Contents of CD */
struct cd_play_track playtrack;             /* Structure to play tracks */ 
struct dir_entry *entry;                    /* Structure of directory */

/* Handy when printing the date */
char months[] = "JanFebMarAprMayJunJulAugSepOctNovDec";

struct {
   char name[20];
   int type,length,track; 
   char size[8];
   char date[7];
   int selected;
   char offset[8];
   long ext_length;
}position[400],list[400],root[400];

struct {
   struct atapimsf msf;
   int state,type;
}track[50];

/*---------------------------------------------------------------------------*/
/* Prototype of all the functions -------------------------------------------*/
/*****************************************************************************/
_PROTOTYPE(void Initialise,       (void));
_PROTOTYPE(int  Abort,            (char *s)); 
_PROTOTYPE(void Getstring,       (int n)); 
_PROTOTYPE(void PrintMsg,         (char *Str,int color)); 
_PROTOTYPE(char PrintQst,         (char *Str,char choice[3])); 
_PROTOTYPE(void WriteCommand,     (char *S));
_PROTOTYPE(char GetAChar,          (void));
_PROTOTYPE(char GetChar,          (void));
_PROTOTYPE(void ReadCommand,      (char *Command));
_PROTOTYPE(void InterpretCommand, (int Command));
_PROTOTYPE(int  execute,          (int) );
_PROTOTYPE(void refresh_list,	  (void));
_PROTOTYPE(void main,             (int argc,char **argv));

_PROTOTYPE (int iso_cmp, (char *name, struct dir_entry *dir_ptr, int dir_flag));
_PROTOTYPE (void list_dir, (struct dir_entry *dir_ptr));
_PROTOTYPE (int copy_file, (void));
_PROTOTYPE (struct dir_entry *look_up, (char *name));
_PROTOTYPE (int read_device, (long offset, int nr_of_bytes, char *buffer)); 
_PROTOTYPE (int valid_fs, (void) );               
_PROTOTYPE (void print_date, (char *date));
_PROTOTYPE (void print_dir_date, (char *date));
_PROTOTYPE (void audio_info, (void));
_PROTOTYPE (void iso_info, (struct iso9660_descriptor *vol_desc));
_PROTOTYPE (void hs_info, (struct high_sierra_descriptor *vol_desc));
_PROTOTYPE (int  iso_711, (char *c));
_PROTOTYPE (int  iso_712, (char *c));
_PROTOTYPE (int  iso_721, (char *c));
_PROTOTYPE (int  iso_722, (char *c));
_PROTOTYPE (int  iso_723, (char *c));
_PROTOTYPE (long iso_731, (char *c));
_PROTOTYPE (long iso_732, (char *c));
_PROTOTYPE (long iso_733, (char *c));

/*----------------------------------------------------------------------------*/
/* Initialisation function ---------------------------------------------------*/
/******************************************************************************/
void Initialise(void)
{
  initscr();
  raw();
  noecho();
}

/*----------------------------------------------------------------------------*/
/* Get the target for the file -----------------------------------------------*/
/******************************************************************************/
void Getstring(int n)
{
   int i,tip=0;
   char c,string[1024];
   for(i=0;i<n;i++)
   {
      c=GetAChar();
      if(c=='\n') break;
      else
         if(c!='?')
            if(c=='#')
            {
               printf("%c %c",8,8);
               c='\0';
               string[tip]='\0';
               tip--;
            }else{
               printf("%c",c);
               string[tip]=c;
               tip++;
            }
   }
   if(string[tip-1]!='/')
   {
      string[tip]='/';
      printf("/");
      tip++;
   }
   string[tip]='\0';
   strcpy(target,string);
}

/*----------------------------------------------------------------------------*/
/* Reset terminal and exit from the program. ---------------------------------*/
/******************************************************************************/
int Abort(char *s)
{
   printf("\x1b[2J\x1b[H\x1b[m");
   move(LINES - 1,0);
   refresh();
   endwin();
   exit(0);
}

/*----------------------------------------------------------------------------*/
/* Prints a message ----------------------------------------------------------*/
/******************************************************************************/
void PrintMsg(char *Str,int color)
{ 
   printf("\x1b[2;45H\x1b[1;37;40m\x1b[0K");
   if (color==RED)
      printf("\x1b[2;45H\x1b[0;31;40m%s",Str);
   if (color==WHITE)
      printf("\x1b[2;45H\x1b[1;37;40m%s",Str);
}

/*----------------------------------------------------------------------------*/
/* Print a Question Message --------------------------------------------------*/
/******************************************************************************/
char PrintQst(char *Str,char choice[3])
{
   char c;

   printf("\x1b[25;2H\x1b[1;37;40m\x1b[0K");
   printf("\x1b[25;2H\x1b[1;37;40m%s ",Str);
   printf("\x1b[1;37;40m(\x1b[1;36;40m%c\x1b[1;37;40m/\x1b[1;36;40m%c\x1b[1;37;40m/\x1b[1;36;40m%c\x1b[1;37;40m) ?",choice[0],choice[1],choice[2]);
   c=0;
   while((LOWER_CASE(c)!=choice[0])&&(LOWER_CASE(c)!=choice[1])&&(LOWER_CASE(c)!=choice[2]))
      c=getch();
   return c;
}

/*----------------------------------------------------------------------------*/
/* Highlights the first letter of S ------------------------------------------*/
/******************************************************************************/
void WriteCommand(char *S)
{
   printf("\x1b[1;36;40m%c\x1b[0;36;40m%s",S[0],S+1);
}

/*----------------------------------------------------------------------------*/
/* Draw Splash Screen --------------------------------------------------------*/
/******************************************************************************/
void Draw_Splash(void)
{
  printf("\x1b[2J");

  printf( "\x1b[7;22H\x1b[0;30;46mͻ");
  printf( "\x1b[8;22H\x1b[0;30;46m   \x1b[1;36;46m A T A P I   ISO 9660 Explorer \x1b[0;30;46m   ");
  printf( "\x1b[9;22H\x1b[0;30;46m                                     ");
  printf("\x1b[10;22H\x1b[0;30;46m      Written  by Alexis LEDOUX      ");
  printf("\x1b[11;22H\x1b[0;30;46m                                     ");
  printf("\x1b[12;22H\x1b[0;30;46m           IT Carlow  1999           "); 
  printf("\x1b[13;22H\x1b[0;30;46m             Version 1.0             ");
  printf("\x1b[14;22H\x1b[0;30;46m                                     ");
  printf("\x1b[15;22H\x1b[0;30;46m      \x1b[1;33;46m<Press a key to continue>\x1b[0;30;46m      ");
  printf("\x1b[16;22H\x1b[0;30;46mͼ\x1b[0;37;40m");
}

/*----------------------------------------------------------------------------*/
/* Draw user interface -------------------------------------------------------*/
/******************************************************************************/
void Draw_interface(void)
{
  int i;
  printf("\x1b[2J");

  printf("\x1b[1;2H\x1b[0;30;46mͻ");
  printf("\x1b[2;2H\x1b[0;30;46m   \x1b[1;36;46m I S O 9660  E X P L O R E R  ");
  printf("\x1b[0;30;46m   ");
  printf("\x1b[3;2H\x1b[30;46mĶ");
  printf("\x1b[4;2H");
  for(i=0;i<19;i++) printf("\x1b[30;46m                                  \x1b[1B\x1b[38D");
  printf("\x1b[23;2H\x1b[30;46mͼ\x1b[0;37;40m"); 
  printf("\x1b[8;42H\x1b[0;30;46mͻ");
  printf("\x1b[9;42H\x1b[30;46m        ");
  printf("\x1b[1;36;46mI N F O R M A T I O N\x1b[0;30;46m        \x1b[10;42H");
  for(i=0;i<13;i++)
     printf("\x1b[30;46m                                     \x1b[1B\x1b[39D");
  printf("\x1b[23;42H\x1b[30;46mͼ\x1b[0;37;40m");
  
  printf("\x1b[4;45H");  WriteCommand("Eject");
  printf("\x1b[4;62H");  WriteCommand("Refresh");
  printf("\x1b[5;45H");	 WriteCommand("Copy file to ...");
  printf("\x1b[5;62H");	 WriteCommand("Minix style ...");
  printf("\x1b[6;45H");  WriteCommand("Quit"); 
  printf("\x1b[24;2H");  WriteCommand("Source :");
  printf("\x1b[25;2H");  WriteCommand("Target :"); 
}
/*---------------------------------------------------------------------------*/
/* Get a character from the keyboard ----------------------------------------*/
/*****************************************************************************/ 
char GetAChar(void)
{
  int c;
  c = getch();
  if (c < 0) abort();
  if (c == '\033') {	/* arrow key */
     if ((c = getch()) == '[') {
	c = getch();
	switch (c) {
	   case 'A': c='?'; break;
	   case 'B': c='?'; break;
	   case 'C': c='?'; break;
	   case 'D': c='?'; break;
	   default:  c='?'; break;
	}
     }	
  else
     c = '?';
  } 
  if(c==13) c='\n'; 
  if(c==8) c='#'; 
  if (isupper(c)) return tolower(c); 
  else return c;
}

/*---------------------------------------------------------------------------*/
/* Get a character from the keyboard ----------------------------------------*/
/*****************************************************************************/ 
char GetChar(void)
{
  int c;

  c = getch();
  if (c < 0) abort();
  if (c == '\033') {	/* arrow key */
     if ((c = getch()) == '[') {
	c = getch();
	switch (c) {
	   case 'A': c = 'U'; break;
	   case 'B': c = 'B'; break;
	   case 'C': c = 'I'; break;
	   case 'D': c = 'L'; break;
	   default:  c = '?'; break;
	}
     }	
  else
     c = '?';
  } 
  if (islower(c)) return toupper(c); 
  else return c;
}

/*----------------------------------------------------------------------------*/
/* Print the name of the file to the given position --------------------------*/
/******************************************************************************/
void PrintPos(int post,int st)
{
   int i,time;

   switch(post)
   {
      case 0: printf("\x1b[4;3H");break;
      case 1: printf("\x1b[5;3H");break;
      case 2: printf("\x1b[6;3H");break;
      case 3: printf("\x1b[7;3H");break;
      case 4: printf("\x1b[8;3H");break;
      case 5: printf("\x1b[9;3H");break;
      case 6: printf("\x1b[10;3H");break;
      case 7: printf("\x1b[11;3H");break;
      case 8: printf("\x1b[12;3H");break;
      case 9: printf("\x1b[13;3H");break;
      case 10: printf("\x1b[14;3H");break;
      case 11: printf("\x1b[15;3H");break;
      case 12: printf("\x1b[16;3H");break;
      case 13: printf("\x1b[17;3H");break;
      case 14: printf("\x1b[18;3H");break;
      case 15: printf("\x1b[19;3H");break;
      case 16: printf("\x1b[20;3H");break;
      case 17: printf("\x1b[21;3H");break;
      case 18: printf("\x1b[22;3H");break;
      case 19: printf("\x1b[4;22H");break;
      case 20: printf("\x1b[5;22H");break;
      case 21: printf("\x1b[6;22H");break;
      case 22: printf("\x1b[7;22H");break;
      case 23: printf("\x1b[8;22H");break;
      case 24: printf("\x1b[9;22H");break;
      case 25: printf("\x1b[10;22H");break;
      case 26: printf("\x1b[11;22H");break;
      case 27: printf("\x1b[12;22H");break;
      case 28: printf("\x1b[13;22H");break;
      case 29: printf("\x1b[14;22H");break;
      case 30: printf("\x1b[15;22H");break;
      case 31: printf("\x1b[16;22H");break;
      case 32: printf("\x1b[17;22H");break;
      case 33: printf("\x1b[18;22H");break;
      case 34: printf("\x1b[19;22H");break;
      case 35: printf("\x1b[20;22H");break;
      case 36: printf("\x1b[21;22H");break;
      case 37: printf("\x1b[22;22H");break;
      default: return;
   }
   if (position[startlist+post].name[0]=='\0'&&position[startlist+post].type!=CDAUDIO)
   {
      if(st==NEW)
         printf("\x1b[0;35;45m%-17s\x1b[0;30;46m","0");

      if(st==OLD)
         printf("\x1b[0;36;46m%-17s\x1b[0;30;46m","0");
      return;
   }
   if (st==OLD)
   {
      if(position[startlist+post].type==DIRECTORY)
         printf("\x1b[0;34;46m %c%-13s%c \x1b[0;30;46m",'[',position[startlist+post].name,']');
      else if((position[startlist+post].type==NOTDIRECTORY)&&(position[startlist+post].selected==TRUE))
              printf("\x1b[1;33;46m<\x1b[0;30;46m%-15s\x1b[1;33;46m>\x1b[0;30;46m",position[startlist+post].name); 
      else if((position[startlist+post].type==NOTDIRECTORY)&&(position[startlist+post].selected==FALSE))
              printf("\x1b[0;30;46m %-15s \x1b[0;30;46m",position[startlist+post].name); 
      else if(position[startlist+post].type==CDAUDIO) printf("\x1b[1;34;46m Audio Track  %02d \x1b[0;30;46m",position[startlist+post].track); 
   }
   if (st==NEW)
   {
      if(position[startlist+post].type==DIRECTORY)
         printf("\x1b[1;36;45m %c%-13s%c \x1b[0;30;46m",'[',position[startlist+post].name,']');
      else if((position[startlist+post].type==NOTDIRECTORY)&&(position[startlist+post].selected==TRUE))
              printf("\x1b[1;36;45m<%-15s>\x1b[0;30;46m",position[startlist+post].name); 
      else if((position[startlist+post].type==NOTDIRECTORY)&&(position[startlist+post].selected==FALSE))
              printf("\x1b[1;36;45m %-15s \x1b[0;30;46m",position[startlist+post].name); 
      else if(position[startlist+post].type==CDAUDIO) printf("\x1b[1;36;45m Audio Track  %02d \x1b[0;30;46m",position[startlist+post].track);
   }
   if (position[startlist+post].type==CDAUDIO)
   {
      i=position[startlist+post].track;
      time=(track[i].msf.Min*60+track[i].msf.Sec)-(track[i-1].msf.Min*60+track[i-1].msf.Sec);
      printf("\x1b[30;46m\x1b[21;44HTime : %02d:%02d\x1b[0;36;46m%8s",time/60,time%60,"0");
      printf("\x1b[36;46m\x1b[22;44H%31s","0");
   }else{ 
      if (!(position[startlist+post].name[1]=='.'&&position[startlist+post].type==DIRECTORY))
      {
         printf("\x1b[0;30;46m\x1b[21;44HSize : %9ld Bytes",iso_733(position[startlist+post].size));
         printf("\x1b[0;30;46m\x1b[22;44HDate : "); print_dir_date(position[startlist+post].date);
      }
      printf("\x1b[0;30;46m\x1b[20;44H%30s"," ");
      if(nb>0) printf("\x1b[1;33;46m\x1b[20;44H%d File%c %d Bytes",nb,nb>1?'s':' ',totsize);
   }
}

/*----------------------------------------------------------------------------*/
/* Reads in a valid command character ----------------------------------------*/
/******************************************************************************/
void ReadCommand(char *Command)
{
  int ValidCommand;

  do {
	ValidCommand = TRUE;

	PrintPos(oldpos,OLD);
        PrintPos(pos,NEW); 

        *Command = GetChar();	/* Read from keyboard */
	switch (*Command) {
	    case '\n':		/* '\n', '\r' or space means place a */
	    case ' ' :
               break;		/* stone at the cursor position  */

	    case 'E': PrintMsg("Eject the CD",WHITE);break; 
	    case 'R': PrintMsg("Refresh",WHITE);break; 
	    case 'Q': PrintMsg("Bye Bye !!",WHITE);break;
	    case 'M': PrintMsg("Go to DOS text files Style",WHITE);break;
	    case 'D': PrintMsg("Go to MINIX text files Style",WHITE);break;

	    /* POSITION */
	    case 'L':
	    case 'I':
	    case 'U':
	    case 'B': break;
	    case '8': *Command = 'U'; break;
	    case '2': *Command = 'B'; break;
	    case '4': *Command = 'L'; break;
	    case '6': *Command = 'I'; break;

	    default:break;
       }
   } while (!ValidCommand);
}

/*----------------------------------------------------------------------------*/
/* Interpret the command -----------------------------------------------------*/
/******************************************************************************/
void InterpretCommand(Command)
char Command;
{
  int Temp,i,sec;
  char c;

  printf("\x1b[2;45H\x1b[1;37;40m\x1b[0K");
  printf("\x1b[2;45H\x1b[1;37;40mUse Arrow key to move and Space");

  switch (Command) {
      /* LEFT */
      case 'L':
        oldpos=pos;
        if (pos>18) pos-=19;
        else
        {
           if (startlist==0) pos=0;
           else
           { startlist-=19;
             for(i=0;i<38;i++) PrintPos(i,OLD); 
           }
        } 
        break;

      /* RIGHT */
      case 'I':
        oldpos=pos; 
        if ((pos+startlist+18)<(nr_files+nr_dir+nr_audio-1))
        {
           if(pos<19) pos+=19;
           else
           {
              startlist+=19;
              pos-=19;
              for(i=0;i<38;i++) PrintPos(i,OLD); 
           }
        }else{
           if((startlist+19)<nr_files+nr_dir+nr_audio)
           { 
              startlist+=19;
              pos-=19; 
              if (pos<0)
              {
                 pos+=19;
                 pos+=(nr_files+nr_dir+nr_audio)-(startlist+pos);
              }
              for(i=0;i<38;i++) PrintPos(i,OLD); 
           }else{
              pos+=(nr_files+nr_dir+nr_audio)-(startlist+pos);
           }
        }

	break;
      /* DOWN */
      case 'B':
        oldpos=pos;
        if ((startlist+pos)<(nr_files+nr_dir+nr_audio))
        {
           if ((startlist>=0)&&(pos==37))
           {
              if((startlist+19)>nr_files+nr_dir+nr_audio) pos=nr_audio+nr_files+nr_dir-startlist;
              else startlist+=19;
              pos=18;
              for(i=0;i<38;i++) PrintPos(i,OLD); 
           }
           if (pos<NR_ELT_PER_COL*NR_OF_COLS-1) pos++;
        } 
        break;
      /* UP */
      case 'U':
        oldpos=pos;
        if ((pos+startlist)>0)
        {
           if((startlist>0)&&(pos==0))
           {
              if((startlist-19)<0) pos=0;
              else startlist-=19; 
              pos=17;
              for(i=0;i<38;i++) PrintPos(i,OLD); 
           }
           if (pos>0) pos--;
        }

        break;

      /* SPACE PRESSED */
      case ' ':
        if (position[startlist+pos].type==CDAUDIO)
        {
           if (track[position[startlist+pos].track].state==CDPLAY)
           {
              execute(CDIOSTOP);
              track[position[startlist+pos].track].state=CDSTOP;
           }else{ 
              printf("\x1b[16;44H\x1b[0;30;46mPlaying Track %02d",position[startlist+pos].track);
              playtrack.end_track=playtrack.begin_track=position[startlist+pos].track; 
              execute(CDIOPLAYTI);
              for(i=0;i<50;i++) track[i].state=CDSTOP;
              track[position[startlist+pos].track].state=CDPLAY;
           }
           break;
        }

        if (position[startlist+pos].type==NOTDIRECTORY)
        {
           if (position[startlist+pos].selected==TRUE)
           {
              position[startlist+pos].selected=FALSE;
              totsize-=iso_733(position[startlist+pos].size);
              nb--;
           }else{
              position[startlist+pos].selected=TRUE;
              totsize+=iso_733(position[startlist+pos].size);
              nb++;
           }
           break;
        }

        if (!(position[startlist+pos].name[0]=='.' && position[startlist+pos].name[1]=='\0'))
        {
           if ((position[startlist+pos].name[0]=='.') && (position[startlist+pos].name[1]=='.'))
           {  
              for(i=path_length;path[i]!='/';i--)
              {
                 path[i]='\0';
                 path_length--;
              }
              if (path_length>0) path[i]='\0';
              path_length++;
              if(path[1]=='\0')
              {
                 for(i=0;i<400;i++)
                 {  
                    memcpy(list[i].name,root[i].name,20);
                    list[i].type=root[i].type;
                    memcpy(list[i].size,root[i].size,8);
                    memcpy(list[i].date,root[i].date,7);
                    list[i].length=root[i].length;
                    list[i].selected=root[i].selected;
                    memcpy(list[i].offset,root[i].offset,8);
                    list[i].ext_length=root[i].ext_length;  
                 }
                 pos=0;
              }
              pos=0;
           }
           else
           {
              if (path[1]!='\0')
              {
                 path[path_length-1]='/';
                 for(i=0;position[startlist+pos].name[i] != '\0';i++) path[path_length+i]=position[startlist+pos].name[i];
                 path[path_length+i]='\0';
              }else{
                 for(i=0;position[startlist+pos].name[i] != '\0';i++) path[i+1]=position[startlist+pos].name[i];
                 path[i+1]='\0';
              }
              path_length+=i;
              path_length++;
           }

           nb=0;
           totsize=0;

           printf("\x1b[24;11H\x1b[1;36;40m%-40s",path);

           PrintMsg("List ...",WHITE);
           if (path[1]!='\0') list_dir(entry);
           pos=0;
           refresh_list();

           printf("\x1b[24;11H\x1b[1;36;40m%-40s",path);
        }
        break;
      case 'M':
           printf("\x1b[5;62H");  WriteCommand("Dos style ...");
           Aflag=0; break;
      case 'D':
           printf("\x1b[5;62H");  WriteCommand("Minix style ...");
           Aflag=1; break;
           
      /* Copy selected file */
      case 'C':
        {
           if(position[startlist+pos].type==NOTDIRECTORY) 
              c=PrintQst("Copy Selected files, the selected File or Abort","sfa");
           if(position[startlist+pos].type==DIRECTORY)
              c=PrintQst("Copy Selected files, the Directory or Abort","sda");  
           printf("\x1b[25;2H\x1b[0K");
           printf("\x1b[25;2H");
           printf("\x1b[1;36;40m%c\x1b[0;36;40m%s",'T',"arget :");
           printf("\x1b[25;11H\x1b[1;36;40m");
           if(c=='a') break; 
           if((c=='s')&&(nb==0)) break;
           Getstring(65);
           if ((c=='s')&&(nb>0))
           { 
              char *ctemp;

              Temp=startlist;
              sec=pos;
              startlist=0; 
              strcpy(ctemp,target);
              for(pos=0;pos<400;pos++)
              {  
                 if(position[pos].name[0]=='\0') break;
                 if(position[pos].selected==TRUE)
                 {   
                    copy_file(); 
                    strcpy(target,ctemp);
                 }
              }
              startlist=Temp;
              pos=sec;
              memset(target,'\0',80);
              printf("\x1b[2;45H\x1b[1;37;40m\x1b[0K");
              printf("\x1b[2;45H\x1b[1;37;40mSelection Saved Successfully"); 
              break;
           }
           if ((c=='f')&&(position[startlist+pos].type==NOTDIRECTORY))
           { 
              if(copy_file()==-1)
              {
                 PrintMsg("File Not Saved",RED);

              }
              else PrintMsg("File Saved Successfully",WHITE);
           } 
           if ((c=='d')&&(position[startlist+pos].type==DIRECTORY))
           {   
              char *ctemp;
              strcat(target,position[startlist+pos].name);
              strcat(target,"/");
              strcpy(ctemp,target);
              if(mkdir(target,755)==-1)
              {
                 PrintMsg("Directory Not Saved",RED);
                 break;
              }else{ 
                 Temp=startlist;
                 sec=pos;
                 list_dir(entry);
                 refresh_list();
                 startlist=0; 
                 for(pos=0;pos<400;pos++)
                 {
                    if(position[pos].name[0]=='\0') break;
                    copy_file(); 
                    strcpy(target,ctemp);
                 } 
                 startlist=Temp;
                 pos=sec;
                 memset(target,'\0',80);
                 PrintMsg("Directory Saved Successfully",WHITE);
              }
           }
           break;
        }
      /* Eject the CDROM */
      case 'E':
         execute(CDIOSTOP);
         execute(CDIOEJECT);
         break;
      /* Read the playlist */
      case 'Q':
         execute(CDIOSTOP);
         break;
      case 'R':
         nr_files=0;
         nr_dir=0;
         nb=0;
         totsize=0;
        
	 if (!valid_fs())
	 {
	    PrintMsg("File system not ISO9660",RED);
	    break;
	 }
         if (!AudioCd)
         {
            if (Iso9660) 
               iso_info(Iso_Vol_Desc);
            else 
               hs_info(Hs_Vol_Desc);
            startlist=0;

            path_length=1;
            path[0] = '/';
            path[1] = '\0';
            Read_Dir=1;
          
            if((entry = look_up(path)) != NULL_DIR)
            {
               list_dir(entry);
               refresh_list();
            }
            printf("\x1b[24;11H\x1b[1;36;40m%-40s",path);
         }else{
            audio_info();
            refresh_list();
         }
         break;
   }
}

/*---------------------------------------------------------------------------*/
/* Execute the IOCTL Command ------------------------------------------------*/
/*****************************************************************************/
int execute(int command)
{ 
   int status;
   switch(command)
   {
      case CDIOSTOP    : status=ioctl(fd,CDIOSTOP,0);
                         break;
      case CDIOEJECT   : status=ioctl(fd,CDIOEJECT,0);
                         break;
      case CDIOPLAYTI  : status=ioctl(fd,CDIOPLAYTI,&playtrack);
      		         break;
      case CDIOCHKSTATUS : status=ioctl(fd,CDIOPLAYTI,&playtrack);
      		         break;
   }
   return status;
}

/*---------------------------------------------------------------------------*/
/* Refresh the list ---------------------------------------------------------*/
/*****************************************************************************/
void refresh_list(void)
{
   int i,j=0,k;

   for(i=0;i<400;i++)
      if(list[i].type==DIRECTORY)
      { 
         memcpy(position[j].name,list[i].name,20);
         position[j].type=DIRECTORY;
         position[j].selected=list[i].selected;
         memcpy(position[j].offset,list[i].offset,8);
         position[j].length=list[i].length;
         position[j].ext_length=list[i].ext_length;  
         memcpy(position[j].size,list[i].size,8);
         memcpy(position[j].date,list[i].date,7);  
         PrintPos(j,OLD);
         j++;
      }
   nr_dir=j-1;
   for(i=0;i<400;i++)
      if(list[i].type==NOTDIRECTORY)
      { 
         memcpy(position[j].name,list[i].name,20);
         position[j].type=NOTDIRECTORY;
         position[j].selected=list[i].selected;
         memcpy(position[j].offset,list[i].offset,8);
         position[j].length=list[i].length;
         position[j].ext_length=list[i].ext_length;  
         memcpy(position[j].size,list[i].size,8);
         memcpy(position[j].date,list[i].date,7);  
         PrintPos(j,OLD);
         j++;
      }
   nr_files=j-nr_dir-1;
   for(i=0;i<50;i++)
      if(track[i].type==CDAUDIO)
      { 
         memset(position[j].name,'\0',20); 
         position[j].type=CDAUDIO;  
         position[j].track=i;
         PrintPos(j,OLD);
         j++;
      }
   nr_audio=j-nr_files-nr_dir-1;
   for(k=j;k<400;k++)
   { 
      memset(position[k].name,'\0',20);
      position[k].type=5;
      PrintPos(k,OLD);
   }

   if (j>0)
   {
      printf("\x1b[19;44H\x1b[36;46m%35s","0");
      printf("\x1b[19;44H\x1b[30;46m%d Director%-3s %d File%s %d Track%s"
             ,nr_dir+1,nr_dir+1>1?"ies":"y",nr_files,nr_files>1?"s":"",nr_audio,
             nr_audio>1?"s":""); 
   }
}

/*----------------------------------------------------------------------------*/
/* Main function -------------------------------------------------------------*/
/******************************************************************************/
void main(int argc,char **argv)
{
  char *device;
  char Command;
  
  /* Copy the value of the first argument in the device variable */
  if (argc==2) device=argv[1];
  else device="/dev/cd1";

  /* Initialise display and variables and Draw the interface */
  Initialise();
  Draw_Splash();
  getch(); 
  Draw_interface();

  /* Initialise ATAPI CDROM */
  fd=open(device,O_RDONLY);

  PrintMsg("Welcome to ISO 9660 Explorer",WHITE);

  /* Interpret commands until the command 'Quit' */
  do{
        ReadCommand(&Command);
        InterpretCommand(Command);
  } while (Command != 'Q');

  /* Close playlist file */
  Abort("Good bye!");
}

/******************************************************************************/
/* ------------------------ ISO 9660 Functions -------------------------------*/
/******************************************************************************/

/*----------------------------------------------------------------------------*/
/* Compare name with directory entries ---------------------------------------*/
/******************************************************************************/
int iso_cmp(char *name, struct dir_entry *dir_ptr, int dir_flag)
{
/* Compare name with directory entries, looking for match with a dirname.
 * An iso9660 filename is terminated by ";n", where n will probably 
 * be 1. A directory name is not terminated by anything special, it may be
 * followed by a \0 if padding is needed to put the following directory 
 * entry on an even address.
 */
  int i;
  int len;

  /* First match the filename */
  len = strlen(name);
  if (len > iso_711(dir_ptr->name_length)) return 1;
  for (i = 0; i < len; i++)
  { 
    if (dir_ptr->name[i] == ';') return 1;	/* found end of a filename */ 
    if (name[i] != LOWER_CASE(dir_ptr->name[i])) return 1; /* match failed */
  }
  if (dir_ptr->name[i] != ';' && i != len) return 1; /* incomplete match */

  /* The filename is ok, now look at the file type */
  if (dir_flag && !IS_DIR(dir_ptr)) return 1;  /* File type not correct */

  return 0; 
}

/*----------------------------------------------------------------------------*/
/* Lookup a file name --------------------------------------------------------*/
/******************************************************************************/
struct dir_entry *look_up(char *path)
{
  /* Lookup a file name */

  struct dir_entry *dir_ptr;
  long block;
  int nr_of_blocks,offset;
  char name[MAX_NAME_LENGTH + 1];
  int name_index = 0;
  int last_in_path = 0;
  int found,i,j; 

  /* Get the right dir entry structure */
  if (Iso9660)
    dir_ptr = (struct dir_entry *) Iso_Vol_Desc->root_dir_entry;  
  else
    dir_ptr = (struct dir_entry *) Hs_Vol_Desc->root_dir_entry;  

  /* If we look for the root we already have the right entry */
  if (path[0] == '/')
    if (strlen(path) == 1) return dir_ptr;
    else name_index = 1; /* first name in path */

  /* Keep searching for the path elements until all are found */
  while (!last_in_path)
  {
    /* Get next name in path */ 
    for (i = name_index; i < strlen(path); i++)
    {
      if (path[i] == '/') break;
      name[i - name_index] = path[i];
    }

    last_in_path = 
           (i == strlen(path) || (i == strlen(path) - 1 && path[i] == '/'));
    name[i-name_index] = '\0';
    name_index = i + 1;
   
    /* Get block of next directory */
    block = iso_733(dir_ptr->first_block) + iso_711(dir_ptr->ext_attr_length);
    nr_of_blocks = (int) (iso_733(dir_ptr->size) >> BLOCK_SHIFT);

    /* Search for file in dir entry */
    found = 0;
    for (j=0; j < nr_of_blocks && !found; j++) 
    {
      /* Read a directory block */
      read_device(block, BLOCK_SIZE, Buffer);

      dir_ptr = (struct dir_entry *) Buffer;

      offset = 0;
      /* Compare with all entries in this block */
      while (iso_711(dir_ptr->length) > 0 && offset < BLOCK_SIZE)
      {
        if (iso_cmp(name, dir_ptr,
            (Read_Dir || (!Read_Dir && !last_in_path))) == 0) 
        {
          found = 1;
          break;
        }
        /* Next entry */
        offset += iso_711(dir_ptr->length);
        dir_ptr = (struct dir_entry *) (Buffer + offset);
      }
    }
    if (!found) return NULL_DIR;   /* path element not found */ 
  }
  return dir_ptr;
}

/*----------------------------------------------------------------------------*/
/* List all entries in a directory and Audio Tracks --------------------------*/
/******************************************************************************/
void list_dir(struct dir_entry *dir_ptr)
{
  /* List all entries in a directory */

  long block;
  int nr_of_blocks;
  int i,j,name_len;
  int offset = 0;
  char *name;
  int post=0; 

  /* Get first block of directory */
  if ((nr_files==0)&&(nr_dir==0))
  {
     block = iso_733(dir_ptr->first_block) + iso_711(dir_ptr->ext_attr_length);
     nr_of_blocks = (int) (iso_733(dir_ptr->size) >> BLOCK_SHIFT);
  }else{
     block = iso_733(position[startlist+pos].offset)+position[startlist+pos].ext_length;
     nr_of_blocks = (int) (iso_733(position[startlist+pos].size) >> BLOCK_SHIFT);
  }  

  memset(list,'\0',sizeof(list));

  /* Read all directory blocks and display their contents */
  for (j=0; j < nr_of_blocks; j++) 
  {
    read_device(block, BLOCK_SIZE, Buffer);
    block++;

    dir_ptr = (struct dir_entry *) (Buffer);
    offset = 0;
    while (iso_711(dir_ptr->length) != 0 && offset < BLOCK_SIZE)
    {
      name_len = 0;
      if (IS_DOT(dir_ptr)) name_len++;
      else
      {
        if (IS_DOT_DOT(dir_ptr))
        { 
          name_len+=2;
          list[post].name[0]='.';
          list[post].name[1]='.';
          list[post].name[2]='\0';
          list[post].type=DIRECTORY; 
          memcpy(list[post].size,dir_ptr->size,8);
          memcpy(list[post].date,dir_ptr->date,7);  
          list[post].length=2;
          list[post].selected=FALSE;
          memcpy(list[post].offset,dir_ptr->first_block,8);
          list[post].ext_length=iso_711(dir_ptr->ext_attr_length);
          if ((nr_files==0)&&(nr_dir==0))
          {
             memcpy(root[post].name,list[post].name,20);
             root[post].type=list[post].type;
             memcpy(root[post].size,list[post].size,8);
             memcpy(root[post].date,list[post].date,7);
             root[post].length=list[post].length;
             root[post].selected=list[post].selected;
             memcpy(root[post].offset,dir_ptr->first_block,8);
             root[post].ext_length=list[post].ext_length;
          }

        }
        else
        {
          for (i=0; i<iso_711(dir_ptr->name_length) && i<NR_OF_CHARS; i++)
          {
             if(dir_ptr->name[i] == ';')break;
             list[post].name[i] = LOWER_CASE(dir_ptr->name[i]);
             name_len++;
          }
          if (IS_DIR(dir_ptr)) list[post].type = DIRECTORY;
          else list[post].type = NOTDIRECTORY;
          list[post].name[i]='\0';
          list[post].length=i;
          list[post].selected=FALSE;
          memcpy(list[post].offset,dir_ptr->first_block,8);
          list[post].ext_length=iso_711(dir_ptr->ext_attr_length);

          memcpy(list[post].size,dir_ptr->size,8);
          memcpy(list[post].date,dir_ptr->date,7);
          if ((nr_files==0)&&(nr_dir==0))
          {
             memcpy(root[post].name,list[post].name,20);
             root[post].type=list[post].type;
             memcpy(root[post].size,list[post].size,8);
             memcpy(root[post].date,list[post].date,7);
             root[post].length=list[post].length;
             root[post].selected=list[post].selected;
             memcpy(root[post].offset,dir_ptr->first_block,8);
             root[post].ext_length=list[post].ext_length;
          }
        }
      }
      post++;
      offset += iso_711(dir_ptr->length);
      dir_ptr = (struct dir_entry *) (Buffer+offset);
    }
  }
}

/*----------------------------------------------------------------------------*/
/* Print Date in a directory entry -------------------------------------------*/
/******************************************************************************/
void print_dir_date(char *date)
{
  /* Print date in a directory entry */

  int i,m;

  m = iso_711(&date[1]) - 1;
  printf("%02d ",date[2]);
  for (i = 0;i<3;i++) printf("%c",months[m*3+i]);

  printf(" 19%02d %02d:%02d:%02d", date[0], date[3], date[4], date[5]);
}

/*----------------------------------------------------------------------------*/
/* List contents of a file ---------------------------------------------------*/
/******************************************************************************/
int copy_file(void)
{
  /* List contents of a file */

  int i;
  long block,size; 
  FILE *target_file;
  struct data_read Data;     /* Packet to retrieve DATA from CD */ 

  block = iso_733(position[startlist+pos].offset);  
  size = iso_733(position[startlist+pos].size);  

  strcat(target,position[startlist+pos].name); 
  target_file=fopen(target,"w");
  if (target_file!=NULL)
  { 
    while (size > 0)
    {
       Data.Offset=block; 
       if(ioctl(fd, CDIOREADDATA, &Data)!=0)
       {
          fclose(target_file);
          unlink(target);
          return -1;
       } 
       for (i=0; ((i < size) && (i < BLOCK_SIZE)); i++)
       {
         if(Data.Data[i] == '\r'){if(Aflag==0)fprintf(target_file,"%c", Data.Data[i]);}
         else fprintf(target_file,"%c", Data.Data[i]);
       }
       size-= BLOCK_SIZE;
       block++;
    } 
    fclose(target_file);
    return 0;
  }
  return -1;
}

/*----------------------------------------------------------------------------*/
/* Print the date in a volume descriptor -------------------------------------*/
/******************************************************************************/
void print_date(char *date)
{  
  printf ("%c%c-%c%c-%c%c%c%c %c%c:%c%c:%c%c",
           date[6],     date[7], date[4],     date[5],
           date[0],     date[1], date[2],     date[3],
           date[8],     date[9], date[10],    date[11],
           date[12],    date[13]);
}

/*----------------------------------------------------------------------------*/
/* Display INFO about AUDIO CD FS --------------------------------------------*/
/******************************************************************************/
void audio_info(void)
{
  int i,time;
  execute(CDIOREADTOC);
  printf("\x1b[11;42H\x1b[0;30;46m Format        : CD-AUDIO            ");
  printf("\x1b[12;42H\x1b[0;30;46m First Track   : %02d                  ",Toc.Hdr.FirstTrk);
  printf("\x1b[13;42H\x1b[0;30;46m Last Track    : %02d                  ",Toc.Hdr.LastTrk);
  time=(track[Toc.Hdr.LastTrk].msf.Min*60+track[Toc.Hdr.LastTrk].msf.Sec);
  printf("\x1b[14;42H\x1b[0;30;46m Total time of : %02d:%02d               ",time/60,time%60);
  printf("\x1b[15;42H\x1b[0;30;46m                                     ");
  printf("\x1b[16;42H\x1b[0;30;46m                                     ");
  printf("\x1b[17;42H\x1b[0;30;46m                                     "); 
}

/*---------------------------------------------------------------------------*/
/* Display Info about ISO9660 FS --------------------------------------------*/
/*****************************************************************************/
void iso_info (struct iso9660_descriptor *vol_desc)
{
  int i;

  printf("\x1b[11;42H\x1b[0;30;46m Format        : ISO9660             ");
  printf("\x1b[12;42H\x1b[30;46m System ID     : ");
  for (i=0; i< 19; i++) printf("%c", vol_desc->system_id[i]);printf(" ");
  printf("\x1b[13;42H\x1b[30;46m               : ");
  for (i=19; i<32; i++) printf("%c", vol_desc->system_id[i]);printf("       ");
  printf("\x1b[14;42H\x1b[30;46m Volume ID     : ");
  for (i=0; i< 19; i++) printf("%c", vol_desc->volume_id[i]);printf(" ");
  printf("\x1b[15;42H\x1b[30;46m               : ");
  for (i=19; i<32; i++) printf("%c", vol_desc->volume_id[i]);printf("       ");
  printf("\x1b[16;42H\x1b[30;46m Volume Size   : %-6ldKb            ",iso_733(vol_desc->volume_size)*2);
  printf("\x1b[17;42H\x1b[30;46m Creation Date : "); print_date(vol_desc->creation_date); 
  printf(" ");
}

/*----------------------------------------------------------------------------*/
/* Display Info about High Sierra FS -----------------------------------------*/
/******************************************************************************/
void hs_info(struct high_sierra_descriptor *vol_desc)
{
  int i;

  printf("\x1b[11;42H\x1b[30;46m Format        : HIGH SIERRA         ");
  printf("\x1b[12;42H\x1b[30;46m System ID     : ");
  for (i=0; i<19; i++) printf("%c", vol_desc->system_id[i]);
  printf(" ");
  printf("\x1b[13;42H\x1b[30;46m               : ");
  for (i=19; i<32; i++) printf("%c", vol_desc->system_id[i]);
  printf("       ");
  printf("\x1b[14;42H\x1b[30;46m Volume ID     : ");
  for (i=0; i< 19; i++) printf("%c", vol_desc->volume_id[i]);
  printf(" ");
  printf("\x1b[15;42H\x1b[30;46m               : ");
  for (i=19; i<32; i++) printf("%c", vol_desc->volume_id[i]);
  printf("       ");
  printf("\x1b[16;42H\x1b[30;46m Volume Size   : %-6ldKb            ",iso_733(vol_desc->volume_size)*2);
}
  
/*----------------------------------------------------------------------------*/
/* Check if it is a valid FS -------------------------------------------------*/
/******************************************************************************/
int valid_fs()               
{
  int i; 

  /* search for a volume descriptor */
  AudioCd=Iso9660=High_Sierra=0;
  for(i=0;i<50;i++) track[i].type=5;
  for(i=0;i<400;i++) list[i].type=5;

  for (i=16; i<100; i++)
  {
    if(read_device((long)i, BLOCK_SIZE, Buffer)==NO)
    {
      if(ioctl(fd,CDIOREADTOC,&Toc)==0)
      { 
         for(i=Toc.Hdr.FirstTrk;i<=Toc.Hdr.LastTrk+1;i++)
            if(((Toc.Ent[i].Adr>>4)==1)&&((Toc.Ent[i].Adr<<4)>>4)==16) 
            {
               track[i].type=CDAUDIO;
               track[i].state=CDSTOP; 
               track[i].msf.Min=Toc.Ent[i].Where.Time.Min;
               track[i].msf.Sec=Toc.Ent[i].Where.Time.Sec;
               track[i].msf.Frame=Toc.Ent[i].Where.Time.Frame;
            }
         AudioCd=1; 
         break;
       }else{
          return 0;
       }
    }

    Iso_Vol_Desc = (struct iso9660_descriptor *) Buffer; 
    Hs_Vol_Desc = (struct high_sierra_descriptor *) Buffer; 
    
    if (strncmp(Iso_Vol_Desc->id, ISO9660_ID, sizeof Iso_Vol_Desc->id) == 0)
    {
       Iso9660 = 1;  
       break;  
    }

    if (strncmp(Hs_Vol_Desc->id, HIGH_SIERRA_ID, sizeof Hs_Vol_Desc->id) == 0)
    {
       High_Sierra = 1;
       break; 
    }
  }
  if(ioctl(fd,CDIOREADTOC,&Toc)==0)
     for(i=Toc.Hdr.FirstTrk;i<=Toc.Hdr.LastTrk;i++)
        if(((Toc.Ent[i].Adr>>4)==1)&&((Toc.Ent[i].Adr<<4)>>4)==16) 
        {
           track[i].msf.Sec=Toc.Ent[i].Where.Time.Sec;
           track[i].msf.Frame=Toc.Ent[i].Where.Time.Frame;
           track[i].msf.Min=Toc.Ent[i].Where.Time.Min;
           track[i].state=CDSTOP;
           track[i].type=CDAUDIO; 
        }

  return (i<100);
}

/*----------------------------------------------------------------------------*/
/* READ Data From the device -------------------------------------------------*/
/******************************************************************************/
int read_device(long offset, int nr_of_bytes, char *buffer)
{
  int i;
  struct data_read Data;     /* Packet to retrieve DATA from CD */ 

  Data.Offset=offset; 
  if(ioctl(fd, CDIOREADDATA, &Data)==0)
  { 
     for(i=0;i<BLOCK_SIZE;i++) 
        Buffer[i]=Data.Data[i];
     return YES;
  }else{
     return NO;
  }
}

/*---------------------------------------------------------------------------*/
/* The ISO9660 functions ----------------------------------------------------*/
/*****************************************************************************/
int iso_711 (char *c) { return (*c & 0xff); }

int iso_712 (char *c)
{
  int n;
	
  n = *c;
  if (n & 0x80) n |= 0xffffff00;
  return n;
}

int iso_721 (char *c) { return ((c[0] & 0xff) | ((c[1] & 0xff) << 8)); }

int iso_722 (char *c) { return (((c[0] & 0xff) << 8) | (c[1] & 0xff)); }

int iso_723 (char *c)
{
  if (c[0] != c[3] || c[1] != c[2]) 
     PrintMsg("Invalid ISO 7.2.3 number",RED);

  return (iso_721 (c));
}

long iso_731 (char *c)
{
  return ((long)(c[0] & 0xff)
       | ((long)(c[1] & 0xff) << 8)
       | ((long)(c[2] & 0xff) << 16)
       | ((long)(c[3] & 0xff) << 24));
}

long iso_732 (char *c)
{
  return (((long)(c[0] & 0xff) << 24)
        | (((long)c[1] & 0xff) << 16)
        | (((long)c[2] & 0xff) << 8)
        | ((long)c[3] & 0xff));
}

long iso_733 (char *c)
{
int i;

  for (i = 0; i < 4; i++) 
     if (c[i] != c[7-i]) PrintMsg("Invalid ISO 7.3.3 number",RED);
     
  return (iso_731(c));
}
