/* This file contains the driver for an Atapi cdrom controller.
 *
 * The file contains one entry point:
 *
 *   mcd_task:		main entry when system is brought up
 *
 *  Mar 12 1999			Author: Alexis LEDOUX 
 */

/* Header Files */
#include "kernel.h"
#include "driver.h"
#include "drvlib.h"
#include <minix/cdrom.h>
#include <minix/atapi.h>
#include <sys/ioctl.h>
#include <net/hton.h>

#if ENABLE_ATAPI_CDROM

/* Miscellaneous constants. */
#define MAX_TRACKS		104	/* Maximum nr. of tracks */
#define SUB_PER_DRIVE		(NR_PARTITIONS * NR_PARTITIONS)
#define TIMEOUT			32000
#define MCD_BLOCK_SIZE		2048
#define MCD_BLOCK_MASK		2047
#define MCD_BLOCK_SHIFT		11

/* Function prototypes */
FORWARD _PROTOTYPE ( int mcd_open, (struct driver *dp, message *m_ptr));
FORWARD _PROTOTYPE ( int mcd_close, (struct driver *dp, message *m_ptr));
FORWARD _PROTOTYPE ( char *mcd_name, (void));
FORWARD _PROTOTYPE ( struct device *mcd_prepare, (int dev));
FORWARD _PROTOTYPE ( int mcd_schedule, (int proc_nr, struct iorequest_s *iop));
FORWARD _PROTOTYPE ( int mcd_finish, (void));
FORWARD _PROTOTYPE ( void mcd_geometry, (struct partition *entry));
FORWARD _PROTOTYPE ( int c_handler, (int irq));

/* IOCTL Prototypes */
FORWARD _PROTOTYPE ( int mcd_ioctl, (struct driver *dp, message *m_ptr));
FORWARD _PROTOTYPE ( int ioctl_readvol, (message *m_ptr));
FORWARD _PROTOTYPE ( int ioctl_volctrl, (message *m_ptr));
FORWARD _PROTOTYPE ( int ioctl_read_toc, (message *m_ptr));
FORWARD _PROTOTYPE ( int ioctl_play_mss, (message *m_ptr));
FORWARD _PROTOTYPE ( int ioctl_play_ti, (message *m_ptr));
FORWARD _PROTOTYPE ( int ioctl_read_data, (message *m_ptr));
FORWARD _PROTOTYPE ( int ioctl_read_time, (message *m_ptr));

/* ATAPI Commands */
FORWARD _PROTOTYPE ( int   atapiEject, (ubyte Load, ubyte WhichDrive) );
FORWARD _PROTOTYPE ( int   atapi_mode_Sense, (ubyte WhichDrive,int pageno,int modeflag,char *buf,int buflen));
FORWARD _PROTOTYPE ( int   atapi_mode_Select, (ubyte WhichDrive,int pageno, char *buf,int buflen));
FORWARD _PROTOTYPE ( int   atapiPlay, (ubyte WhichDrive, ulong Start, ulong End) );
FORWARD _PROTOTYPE ( int   atapiTOC, (ubyte WhichDrive));
FORWARD _PROTOTYPE ( int   atapiPause, (ubyte Play, ubyte WhichDrive));
FORWARD _PROTOTYPE ( int   atapiStop, (ubyte WhichDrive));
FORWARD _PROTOTYPE ( ulong atapiCapacity, (ubyte WhichDrive));
FORWARD _PROTOTYPE ( int   atapiREAD, (ubyte WhichDrive));
FORWARD _PROTOTYPE ( int   atapiReadTime, (ubyte WhichDrive, ubyte *Data));
FORWARD _PROTOTYPE ( int   atapiStatus, (ubyte WhichDrive) );
FORWARD _PROTOTYPE ( int   atapiLock, (ubyte WhichDrive,ubyte Lockflag) ); 
FORWARD _PROTOTYPE ( int   SetSpeed, (ubyte WhichDrive) );
FORWARD _PROTOTYPE ( int   waitfor, (int mask, int value));
FORWARD _PROTOTYPE ( void  ataProbe, (void) );

/* ATAPI Global variables */
PRIVATE struct ataid DriveInfo;
PRIVATE struct drive Drive[4];              /* Information about the 4 drives */
PRIVATE unsigned NumTracks;                 /* Number of tracks */
PRIVATE struct atapitoc Contents;           /* Table of contents of the CD */
PRIVATE struct atapi_subchnl time;          /* Subchannel structure */
PRIVATE struct volctrl volctrl;             /* Information about the volume */
PRIVATE ulong Capacity;			    /* Capacity total of CDROM */
PRIVATE struct data_read Buffer;            /* Structure to tranfer data */
PRIVATE int CDROM_Drive=-1;

/* Entry points to this driver. */
PRIVATE struct driver atapi_dtab = 
{
  mcd_name,	/* Current device's name */
  mcd_open,	/* Open request read table of contents */
  mcd_close,	/* Release device */
  mcd_ioctl,	/* Do cdrom ioctls */
  mcd_prepare,	/* Prepare for I/O */
  mcd_schedule,	/* Precompute blocks */
  mcd_finish,	/* Do the I/O */
  nop_cleanup,	/* No cleanup to do */
  mcd_geometry	/* Tell geometry */
};

PRIVATE struct trans 
{
  struct iorequest_s *tr_iop;	/* Belongs to this I/O request */
  unsigned long tr_pos;		/* Byte position to transfer from */
  int tr_count;			/* Byte count */
  phys_bytes tr_phys;		/* User physical address */
} mcd_trans[NR_IOREQS];


/* Globals */

PRIVATE int atapi_irq;			/* Interrupt request line */
PRIVATE int atapi_io_base;		/* I/O base register */
PRIVATE struct device *mcd_dv;		/* Active partition */
PRIVATE struct trans *mcd_tp;		/* Pointer to add transfer requests */
PRIVATE struct device mcd_part[DEV_PER_DRIVE];
					/* Primary partitions: cd[0-4] */
PRIVATE struct device mcd_subpart[SUB_PER_DRIVE];
					/* Subpartitions: cd[1-4][a-d] */
PRIVATE int mcd_open_ct;		/* in-use count */
PRIVATE unsigned mcd_count;
PRIVATE unsigned long mcd_nextpos;

/*=========================================================================*
 *				mcd_task				   *
 *=========================================================================*/
PUBLIC void mcd_task()
{
  int i,temp;

  /* Detect if ATAPI CDROM Drive exist */
  ataProbe();

  driver_task(&atapi_dtab);
}

/*=========================================================================*
 *				mcd_open				   *	
 *=========================================================================*/
PRIVATE int mcd_open(dp, m_ptr)
struct driver *dp;	/* pointer to this drive */
message *m_ptr;		/* OPEN */
{
  int i, status;

  if (mcd_prepare(m_ptr->DEVICE) == NIL_DEV) return ENXIO;

  /* A CD-ROM is read-only by definition. */
  if (m_ptr->COUNT & W_BIT) return EACCES;

  if (mcd_open_ct == 0)
  {

    mcd_open_ct++;

    /* fill in size of device (= nr. of bytes on the disk) */
    mcd_part[0].dv_base = 0;

    for(i=0;i<3;i++)
    {
       status=atapiCapacity(CDROM_Drive);
       if (!status) break;
    }

    mcd_part[0].dv_size = (Capacity*4)*2048;

    atapi_io_base = Drive[CDROM_Drive].IOAdr;
    atapi_irq = 15;
    put_irq_handler(atapi_irq, c_handler);
    enable_irq(atapi_irq);

    /* Partition the disk. */
    /*partition(&atapi_dtab, 0, P_PRIMARY);*/
  }
  return OK;
}


/*=========================================================================*
 *				mcd_close				   *	
 *=========================================================================*/
PRIVATE int mcd_close(dp, m_ptr)
struct driver *dp;	/* pointer to this drive */
message *m_ptr;		/* CLOSE */
{
  /* One less reference to the device */

  mcd_open_ct--;
  return OK;
}


/*=========================================================================*
 *				mcd_name				   *	
 *=========================================================================*/
PRIVATE char *mcd_name()
{
  /* Return a name for the device */

  return "Atapi0";
}

/* IOCTL Stuff */

/*=========================================================================*
 *				mcd_ioctl				   *	
 *=========================================================================*/
PRIVATE int mcd_ioctl(dp, m_ptr)
struct driver *dp;	/* pointer to the drive */
message *m_ptr;		/* contains ioctl command */
{
  /* Perform the ioctl request */

  int status,i;

  if (mcd_prepare(m_ptr->DEVICE) == NIL_DEV) return(ENXIO);

  switch(m_ptr->REQUEST)
  {
    case CDIOPLAYMSS:     status=ioctl_play_mss(m_ptr);       break;
    case CDIOPLAYTI:      status=ioctl_play_ti(m_ptr);        break;
    case CDIOREADTOC:     status=ioctl_read_toc(m_ptr);       break;
    case CDIOREADTIME:	  status=ioctl_read_time(m_ptr);      break;
    case CDIOREADDATA:	  status=ioctl_read_data(m_ptr);      break;
    case CDIOVOLCTRL:	  status=ioctl_volctrl(m_ptr);        break;
    case CDIOVOLREAD:	  status=ioctl_readvol(m_ptr);        break;
    case CDIOCHKSTATUS:   status=atapiStatus(CDROM_Drive);    break;
    case CDIOREADCAP:	  status=atapiCapacity(CDROM_Drive);  break;
    case CDIOSTOP:        for(i=0;i<3;i++) 
                             if ((status = atapiStop(CDROM_Drive))==0) break;
                          break;
    case CDIOPAUSE:       for(i=0;i<3;i++) 
                             if ((status = atapiPause(0,CDROM_Drive))==0) break;
                          break;
    case CDIORESUME:      for(i=0;i<3;i++) 
                             if ((status = atapiPause(1,CDROM_Drive))==0) break; 
                          break;
    case CDIOEJECT:       for(i=0;i<3;i++) 
                             if ((status = atapiEject(0,CDROM_Drive))==0) break; 
                          break;
    case CDIOCLOSE:	  for(i=0;i<3;i++)
                             if ((status = atapiEject(1,CDROM_Drive))==0) break; 
                          break;
    case CDIOLOCK:	  for(i=0;i<3;i++)
                             if ((status = atapiLock(CDROM_Drive,1))==0) break;
                          break;
    case CDIOUNLOCK:	  for(i=0;i<3;i++)
                             if ((status = atapiLock(CDROM_Drive,0))==0) break;
                          break;
    default:              status = do_diocntl(dp, m_ptr);
  }
  return status;
}

/*=========================================================================*
 *				ioctl_read_toc	 			   *	
 *=========================================================================*/
PRIVATE int ioctl_read_toc(m_ptr)
message *m_ptr;
{
  phys_bytes user_phys;
  int err, toc_size,i;

  /* Try to read the table of contents */
  for (i=0;i<3;i++)
     if ((err = atapiTOC(CDROM_Drive))==0) break;
  if(err) return err;

  /* Copy to user */
  user_phys = numap(m_ptr->PROC_NR, (vir_bytes) m_ptr->ADDRESS, sizeof(struct atapitoc));
  if (user_phys == 0) return(EFAULT);
  phys_copy(vir2phys(&Contents), user_phys, (phys_bytes) sizeof(struct atapitoc));

  return OK;
}  

/*=========================================================================*
 *				ioctl_play_mss	  			   *	
 *=========================================================================*/
PRIVATE int ioctl_play_mss(m_ptr)
message *m_ptr;
{
  phys_bytes user_phys;
  struct cd_play_mss mss;
  ulong start,end;
  int i,status;

  /* Get user data */
  user_phys = numap(m_ptr->PROC_NR, (vir_bytes) m_ptr->ADDRESS, sizeof(mss));
  if (user_phys == 0) return(EFAULT);
  phys_copy(user_phys, vir2phys(&mss), (phys_bytes) sizeof(mss));

  /* Try to play */
  start=(ulong)((mss.begin_mss[0]*60+mss.begin_mss[1])*75+mss.begin_mss[2]);
  end  =(ulong)((mss.end_mss[0]  *60+mss.end_mss[1])  *75+mss.end_mss[2]);
  for(i=0;i<3;i++)
     if((status=atapiPlay(CDROM_Drive,start,end))==0) break;

  return status;
}

/*=========================================================================*
 *                              ioctl_read_data                            *
 *=========================================================================*/
PRIVATE int ioctl_read_data(m_ptr)
message *m_ptr;
{
  phys_bytes user_phys;
  int i,temp,status;

  /* Get user data */
  user_phys = numap(m_ptr->PROC_NR, (vir_bytes) m_ptr->ADDRESS, sizeof(struct data_read));
  if (user_phys == 0) return(EFAULT);
  phys_copy(user_phys, vir2phys(&Buffer), (phys_bytes) sizeof(struct data_read));

  for(i=0;i<3;i++)
     if((status=atapiREAD(CDROM_Drive))==0) break;
  if(status) return status;

  phys_copy(vir2phys(&Buffer), user_phys,(phys_bytes) sizeof(struct data_read));

  return OK;
}

/*=========================================================================*
 *				ioctl_play_ti	  			   *	
 *=========================================================================*/
PRIVATE int ioctl_play_ti(m_ptr)
message *m_ptr;
{
  phys_bytes user_phys;
  struct cd_play_track tracks;
  int t,i,status;
  ulong start,end;

  /* Get user data */
  user_phys = numap(m_ptr->PROC_NR, (vir_bytes) m_ptr->ADDRESS, sizeof(tracks));
  if (user_phys == 0) return(EFAULT);
  phys_copy(user_phys, vir2phys(&tracks), (phys_bytes) sizeof(tracks));

  /* Try to play */
  t=tracks.begin_track-1;
  start=(ulong)((Contents.Ent[t].Where.Time.Min*60
                +Contents.Ent[t].Where.Time.Sec)
                *75+Contents.Ent[t].Where.Time.Frame);
  t=tracks.end_track; 
  end  =(ulong)((Contents.Ent[t].Where.Time.Min*60
                +Contents.Ent[t].Where.Time.Sec)*75
                +Contents.Ent[t].Where.Time.Frame-1);

  for(i=0;i<3;i++)
     if((status=atapiPlay(CDROM_Drive,start,end))==0) break;

  return status;
}

/******************************************************************************
 *                         ioctl_read_time                                    *
 ******************************************************************************/
PRIVATE int ioctl_read_time(m_ptr)
message *m_ptr;
{
  phys_bytes user_phys;
  int status, i;

  /* Try to read the sub channel */
  for(i=0;i<3;i++)
     if((status=atapiReadTime(CDROM_Drive,(ubyte *)&time))==0) break;
  if (status) return status;

  /* Copy to user */
  user_phys = numap(m_ptr->PROC_NR, (vir_bytes) m_ptr->ADDRESS, sizeof(struct atapi_subchnl));
  if (user_phys == 0) return (EFAULT);
  phys_copy(vir2phys(&time), user_phys, (phys_bytes) sizeof(struct atapi_subchnl));

  return time.ind;
}

/******************************************************************************
 *                           ioctl_readvol                                    *
 ******************************************************************************/
PRIVATE int ioctl_readvol(m_ptr)
message *m_ptr;
{
  phys_bytes user_phys;
  int status,i;
  char buffer[24];

  for(i=0;i<3;i++)
     if((status=atapi_mode_Sense(CDROM_Drive,PAGE_AUDIO,0,buffer,sizeof(buffer)))==0) break;
  if(status) return status;

  volctrl.channel0=buffer[17];
  volctrl.channel1=buffer[19];
  volctrl.channel2=buffer[21];
  volctrl.channel3=buffer[23];

  user_phys = numap(m_ptr->PROC_NR, (vir_bytes) m_ptr->ADDRESS, sizeof(struct volctrl));
  if (user_phys == 0) return (EFAULT);
  phys_copy(vir2phys(&volctrl), user_phys, (phys_bytes) sizeof(struct volctrl));
  return OK;
}

/******************************************************************************
 *                               ioctl_volctrl                                *
 ******************************************************************************/
PRIVATE int ioctl_volctrl(m_ptr)
message *m_ptr;
{
  phys_bytes user_phys;
  int err, subsize;
  char buffer[24],mask[24];

  user_phys = numap(m_ptr->PROC_NR, (vir_bytes) m_ptr->ADDRESS, sizeof(struct volctrl));
  if (user_phys == 0) return (EFAULT);
  phys_copy(user_phys, vir2phys(&volctrl), (phys_bytes) sizeof(struct volctrl));

  err = atapi_mode_Sense(CDROM_Drive,PAGE_AUDIO,0,buffer,sizeof(buffer));
  if(err) return err;

  err = atapi_mode_Sense(CDROM_Drive,PAGE_AUDIO,1,mask,sizeof(buffer));
  if(err) return err;

  buffer[1]=buffer[2]=0;

  buffer[17]=volctrl.channel0 & mask[17];
  buffer[19]=volctrl.channel1 & mask[19];
  buffer[21]=volctrl.channel2 & mask[21];
  buffer[23]=volctrl.channel3 & mask[23];

  err = atapi_mode_Select(CDROM_Drive,PAGE_AUDIO,buffer,sizeof(buffer));

  return err;
}

/* ATAPI CODE */

/*************************************************************************
 *	name:	insw                                                     *
 *	action:	reads Count words (16-bit) from I/O port Adr to          *
 *		memory location Data                                     *
 *************************************************************************/
PRIVATE void insw(uword Adr, uword *Data, uword Count)
{	for(; Count; Count--) *Data++=in_word(Adr);
}

/**************************************************************************
 *	name:	outsw                                                     *
 *	action:	writes Count words (16-bit) from memory location Data     *
 *		to I/O port Adr                                           *
 **************************************************************************/
PRIVATE void outsw(uword Adr, uword *Data, uword Count)
{	for(; Count; Count--) out_word(Adr, *Data++);
}

/*****************************************************************************
 *	name:	ataProbe2                                                        *
 *****************************************************************************/
PRIVATE void ataProbe2(ubyte WhichDrive)
{	ubyte Temp1, Temp2;
	uword IOAdr, Temp;

	IOAdr=Drive[WhichDrive].IOAdr;
	Temp1=in_byte(IOAdr + ATA_REG_CNT);
	Temp2=in_byte(IOAdr + ATA_REG_SECT);
	if(Temp1 != 0x01 || Temp2 != 0x01)
	{
NO_DRIVE:	Drive[WhichDrive].IOAdr=0;
		return;
	}
	Temp1=in_byte(IOAdr + ATA_REG_LOCYL);
	Temp2=in_byte(IOAdr + ATA_REG_HICYL);
	Temp=in_byte(IOAdr + ATA_REG_STAT);

	if(Temp1 == 0x14 && Temp2 == 0xEB)
	{
	        CDROM_Drive=WhichDrive;
		printf("hd%1u (0x%03X, %s): ATAPI ",
		       WhichDrive, IOAdr,Drive[WhichDrive].DrvSel==0xA0 ? "Master" : "Slave"); 
		out_byte(IOAdr + ATA_REG_CMD, ATA_CMD_PID); 
	}else{
	   return;
	}

/* ATAPI: get results of identify */
	milli_delay(200);

/* grab info returned by 'identify' */
	(void)in_byte(IOAdr + ATA_REG_STAT);
	insw(IOAdr + ATA_REG_DATA, (uword *)&DriveInfo,sizeof(DriveInfo) / 2);
	Temp2=1;

	if((DriveInfo.Model[0] == 'N' && DriveInfo.Model[1] == 'E') ||
	  (DriveInfo.Model[0] == 'F' && DriveInfo.Model[1] == 'X') ||
	  (DriveInfo.Model[0] == 'P' && DriveInfo.Model[1] == 'i'))
		Temp2=0;

	for(Temp=0; Temp < 40; Temp += 2)
	{	printf("%c", DriveInfo.Model[Temp ^ Temp2]);
		printf("%c", DriveInfo.Model[Temp ^ Temp2 ^ 1]);
	}
	printf("\n");
}
/*****************************************************************************
	name:	ataSelect
*****************************************************************************/
PRIVATE int ataSelect(uword IOAdr, ubyte Sel)
{	uword Temp;

	Temp=in_byte(IOAdr + ATA_REG_DRVHD);
	if(((Temp ^ Sel) & 0x10) == 0) return(0); /* already selected */
	out_byte(IOAdr + ATA_REG_DRVHD, Sel);

	for(Temp=WAIT_READY; Temp; Temp--)
	{	if((in_byte(IOAdr + ATA_REG_STAT) & 0x80) == 0) break;
		milli_delay(1); }
	return(Temp == 0);
}
/*****************************************************************************
	name:	ataProbe
*****************************************************************************/
PRIVATE void ataProbe(void)
{	ubyte Temp1, Temp2, WhichDrive;
	uword IOAdr, Temp;

/* set initial values */
	Drive[0].DrvSel=Drive[2].DrvSel=0xA0;
	Drive[1].DrvSel=Drive[3].DrvSel=0xB0;
	Drive[0].IOAdr=Drive[1].IOAdr=0x1F0;
	Drive[2].IOAdr=Drive[3].IOAdr=0x170;

	for(WhichDrive=0; WhichDrive < 4; WhichDrive += 2)
	{	IOAdr=Drive[WhichDrive].IOAdr;
/* poke at the interface to see if anything's there */
		out_byte(IOAdr + ATA_REG_CNT, 0x55);
		out_byte(IOAdr + ATA_REG_SECT, 0xAA);
		Temp1=in_byte(IOAdr + ATA_REG_CNT);
		Temp2=in_byte(IOAdr + ATA_REG_SECT);
		if(Temp1 != 0x55 || Temp2 != 0xAA)
/* no master: clobber both master and slave */
NO_DRIVES:	{
			Drive[WhichDrive + 1].IOAdr=Drive[WhichDrive].IOAdr=0;
			continue;
		}
/* soft reset both drives on this I/F (selects master) */
		out_byte(IOAdr + ATA_REG_SLCT, 0x0E);
		milli_delay(200);
/* release soft reset AND enable interrupts from drive */
		out_byte(IOAdr + ATA_REG_SLCT, 0x08);
		milli_delay(200);
/* wait for master */
		for(Temp=2000; Temp; Temp--)	/* XXX - why 2000? */
		{
			if((in_byte(IOAdr + ATA_REG_STAT) & 0x80) == 0) break;
			milli_delay(1);
		}
		if(Temp == 0) goto NO_DRIVES;
		ataProbe2(WhichDrive);
		if(CDROM_Drive>-1) break;
/* select slave */
		if(ataSelect(IOAdr, 0xB0))
/* no response from slave: re-select master and continue */
		{
			ataSelect(IOAdr, 0xA0);
			Drive[WhichDrive + 1].IOAdr=0;
			continue;
		}
		/* identify slave */
		ataProbe2(WhichDrive + 1);
		if(CDROM_Drive>-1) break;
	}
}

/*****************************************************************************
	name:	atapiCmd2
	action:	writes ATA register file including packet command byte,
		busy-waits until drive ready, then writes 12-byte ATAPI
		command	packet, and services interrupts
	returns:0  if success
		-1 if drive could not be selected
		-3 timeout after writing pkt cmd byte (0xA0)
		-4 timeout after writing cmd pkt
		-5 data shortage (premature ATAPI_PH_DONE)
		-6 drive aborted command
		-7 bad drive phase
*****************************************************************************/ 
PRIVATE int atapiCmd2(struct drivecmd *Cmd, ubyte *Pkt)
{	uword IOAdr, Got;
	ubyte Phase;
	ulong Temp;
	int i=0;

	IOAdr=Drive[Cmd->Dev].IOAdr;

	if(ataSelect(IOAdr, Drive[Cmd->Dev].DrvSel)) return(-1); 

	lock();
	   out_byte(IOAdr + ATA_REG_FEAT, 0);	/* irrelevant? */
	   out_byte(IOAdr + ATA_REG_REASON, 0);/* irrelevant? */
	   out_byte(IOAdr + ATA_REG_SECT, 0);	/* irrelevant? */
	   out_byte(IOAdr + ATA_REG_LOCNT, 32768ul);
	   out_byte(IOAdr + ATA_REG_HICNT, 32768ul >> 8);
	   out_byte(IOAdr + ATA_REG_CMD, ATA_CMD_PKT);
	unlock();
	 
	for(Temp=500; Temp; Temp--)
	{	
	   if((in_byte(IOAdr + ATA_REG_STAT) & 0x88) == 0x08) break;
	   milli_delay(1);
	}

	if(Temp == 0) return(-3);
	milli_delay(200);

	   outsw(IOAdr + ATA_REG_DATA, (uword *)Pkt, 6);

	while(1)
	{
	        waitfor(0x80,0);

		Phase=in_byte(IOAdr + ATA_REG_STAT) & 0x08;
                Phase |= (in_byte(IOAdr + ATA_REG_REASON) & 3);

		if(Phase == ATAPI_PH_DONE)
                   if(Cmd->Count) return(-5);  /* could mean no CD or audio CD */
		   else return(0);
		else if(Phase == ATAPI_PH_ABORT) return(-6);
		else if(Phase != ATAPI_PH_DATAIN) return(-7);
	
                Got=in_byte(IOAdr + ATA_REG_HICNT);
		Got=(Got << 8) | in_byte(IOAdr + ATA_REG_LOCNT);
/* Cmd->Count=how many bytes we want to transfer
   Got=how many bytes are available for transfer */
                Temp=MIN(Cmd->Count, Got);

                insw(IOAdr + ATA_REG_DATA, (uword *)Cmd->Data, Temp >> 1);
   
/* read and discard surplus data */
		for(Got -= Temp; Got > 1; Got -= 2) in_word(IOAdr + ATA_REG_DATA);
		if(Got) in_byte(IOAdr + ATA_REG_DATA);
/* advance pointers */
		Cmd->Data += Temp;
		Cmd->Count -= Temp;
		Temp >>= ATAPI_LG_SECTSIZE;
		Cmd->Blk += Temp;
	}
}

/* ATAPI Commands */
/*****************************************************************************
	name:	atapiEject
	action:	loads (Load == 0) or ejects (Load != 0) CD-ROM
	returns:whatever atapiCmd2() returns
*****************************************************************************/
PRIVATE int atapiEject(ubyte Load, ubyte WhichDrive)
{	ubyte Pkt[12]={ATAPI_CMD_START_STOP,0,0,0,0,0,0,0,0,0,0,0};
	struct drivecmd Cmd;

	Cmd.Blk=0;
	Cmd.Count=0;
	Cmd.Dev=WhichDrive;
	Cmd.Cmd=0;
	Cmd.Data=NULL;

	Pkt[4]=2 + (Load != 0);
	return(atapiCmd2(&Cmd, Pkt));
}

PRIVATE int atapiLock(ubyte WhichDrive, ubyte Lockflag)
{
   ubyte Pkt[12]={ATAPI_CMD_LOCK_UNLOCK,0,0,0,0,0,0,0,0,0,0,0};
   struct drivecmd Cmd;

   Cmd.Blk=0;
   Cmd.Count=0;
   Cmd.Dev=WhichDrive;
   Cmd.Cmd=0;
   Cmd.Data=NULL;

   Pkt[4]=(Lockflag!=0);
   return(atapiCmd2(&Cmd, Pkt));
}
/*****************************************************************************
	name:	atapi_mode_Sense
	action:	
	returns:whatever atapiCmd2() returns
*****************************************************************************/
PRIVATE int atapi_mode_Sense(ubyte WhichDrive,int pageno,int modeflag,char* buf,int buflen)
{	ubyte Pkt[12]={ATAPI_MODE_SENSE_10,0,0,0,0,0,0,0,0,0,0,0};
	struct drivecmd Cmd;
	
	Cmd.Blk=0;
	Cmd.Count=buflen;
	Cmd.Dev=WhichDrive;
	Cmd.Cmd=0;
	Cmd.Data=(ubyte *)buf;

	Pkt[2]=pageno|(modeflag<<6);
	Pkt[7]=(buflen>>8);
	Pkt[8]=buflen;

	return(atapiCmd2(&Cmd, Pkt));
}

/*****************************************************************************
	name:	atapi_mode_Select
	action:	
	returns:whatever atapiCmd2() returns
*****************************************************************************/
PRIVATE int atapi_mode_Select(ubyte WhichDrive,int pageno, char* buf, int buflen)
{	ubyte Pkt[12]={ATAPI_MODE_SELECT_10,0,0,0,0,0,0,0,0,0,0,0};
	struct drivecmd Cmd;

	Cmd.Blk=0;
	Cmd.Count=-buflen;
	Cmd.Dev=WhichDrive;
	Cmd.Cmd=0;
	Cmd.Data=(ubyte *)buf;

	Pkt[1]=0x10;
	Pkt[2]=pageno;
	Pkt[7]=(buflen >> 8);
	Pkt[8]=(buflen & 0xff);

	return(atapiCmd2(&Cmd, Pkt));
}

/*****************************************************************************
	name:	atapiPlay
	action:	plays audio from time index Start to End (units of 1/75 sec)
	returns:whatever atapiCmd2() returns
*****************************************************************************/
PRIVATE int atapiPlay(ubyte WhichDrive, ulong Start, ulong End)
{	ubyte Pkt[12]={ATAPI_CMD_PLAY,0,0,0,0,0,0,0,0,0,0,0};
	struct drivecmd Cmd;

	Cmd.Blk=0;
	Cmd.Count=0;
	Cmd.Dev=WhichDrive;
	Cmd.Cmd=0;
	Cmd.Data=NULL;

	Pkt[5]=Start % 75;  Start /= 75;
	Pkt[4]=Start % 60;  Start /= 60;
	Pkt[3]=Start;

	Pkt[8]=End % 75;  End /= 75;
	Pkt[7]=End % 60;  End /= 60;
	Pkt[6]=End;

	return(atapiCmd2(&Cmd, Pkt));
}

/*****************************************************************************
	name:	SetSpeed
	action:	Set the readand write speed
	returns:whatever atapiCmd2() returns
*****************************************************************************/
PRIVATE int SetSpeed(ubyte WhichDrive)
{	ubyte Pkt[12]={ATAPI_CMD_SPEED,0,0,0,0,0,0,0,0,0,0,0};
	struct drivecmd Cmd;
	int speed;
	speed=24*177;

	Cmd.Blk=0;
	Cmd.Count=0;
	Cmd.Dev=WhichDrive;
	Cmd.Cmd=0;
	Cmd.Data=NULL;

	Pkt[2]=speed>>8;
	Pkt[3]=speed;

	return(atapiCmd2(&Cmd, Pkt));
}
	
/*****************************************************************************
	name:	atapiTOCEnt
	action:	reads one or more table-of-contents entries from audio CD
	returns:whatever atapiCmd2() returns
*****************************************************************************/
PRIVATE int atapiTOCEnt(ubyte *Contents, uword Count, ubyte WhichDrive)
{	ubyte Pkt[12]={ATAPI_CMD_READTOC,0,0,0,0,0,0,0,0,0,0,0};
	struct drivecmd Cmd;

	Cmd.Blk=0;
	Cmd.Count=Count;
	Cmd.Dev=WhichDrive;
	Cmd.Cmd=0;
	Cmd.Data=Contents;

	Pkt[1]=2;
	Pkt[6]=0;
	Pkt[7]=Count >> 8;
	Pkt[8]=Count;
	Pkt[9]=0;
	return(atapiCmd2(&Cmd, Pkt));
}

/*****************************************************************************
	name:	atapiREAD
	action:	reads one sector from DATA CD
	returns:whatever atapiCmd2() returns
*****************************************************************************/
PRIVATE int atapiREAD(ubyte WhichDrive)
{	struct drivecmd Cmd;
	ubyte Pkt[12]={ATAPI_CMD_READ10,0,0,0,0,0,0,0,0,0,0,0};
	int format;
	long tmp;

        Cmd.Blk=Buffer.Offset;
        Cmd.Count=2048;
	Cmd.Dev=WhichDrive;
	Cmd.Cmd=DRV_CMD_RD;
	Cmd.Data=(ubyte *)Buffer.Data;

	format=2;
        Pkt[1]=0;

        tmp=Buffer.Offset;
        Pkt[2]=tmp>>24;
        Pkt[3]=tmp>>16;
        Pkt[4]=tmp>>8;
        Pkt[5]=tmp;

        tmp=2048;
        Pkt[6]=tmp>>16;
        Pkt[7]=tmp>>8;
        Pkt[8]=tmp;
        Pkt[9]=0;
        Pkt[10]=Pkt[11]=0;
	return(atapiCmd2(&Cmd, Pkt));
}

/*****************************************************************************
	name:	atapiTOC
	action:	reads table of contents of audio CD and prints starting
		time of each track
	returns:whatever atapiCmd2() returns
*****************************************************************************/
PRIVATE int atapiTOC(ubyte WhichDrive)
{	int Temp,Temp1;

/* read just the header at first */

	Temp1=sizeof(struct atapitocheader);
        Temp=atapiTOCEnt((ubyte *)&Contents.Hdr, Temp1, WhichDrive);
        if(Temp) return(Temp);
        NumTracks=Contents.Hdr.LastTrk - Contents.Hdr.FirstTrk + 1;
        if(NumTracks <= 0 || NumTracks > 99)
	{	printf("atapiTOC: bad number of tracks %d\n", NumTracks);
		return(-1); }
	if(NumTracks > MAX_TRACKS)
	{	printf("Too many tracks (%u); reducing to %u.\n", NumTracks, MAX_TRACKS);
		NumTracks=MAX_TRACKS; }
	Temp1=sizeof(struct atapitocheader) + (NumTracks + 1) * sizeof(struct atapitocentry);
        Temp=atapiTOCEnt((ubyte *)&Contents.Hdr, Temp1, WhichDrive);
	if(Temp) return(Temp); 
        return(0);
}
/*****************************************************************************
	name:	atapiPause
	action:	pauses (Play == 0) or continues (Play != 0) audio CD
	returns:whatever atapiCmd2() returns
*****************************************************************************/
PRIVATE int atapiPause(ubyte Play, ubyte WhichDrive)
{	ubyte Pkt[12]={
		ATAPI_CMD_PAUSE,0,0,0,0,0,0,0,0,0,0,0};
	struct drivecmd Cmd;

	Cmd.Blk=0;
	Cmd.Count=0;
	Cmd.Dev=WhichDrive;
	Cmd.Cmd=0;
	Cmd.Data=NULL;

	Pkt[8]=(Play != 0);
	return(atapiCmd2(&Cmd, Pkt));
}
/*****************************************************************************
	name:	atapiStop
	action:	Stop audio CD
	returns:whatever atapiCmd2() returns
*****************************************************************************/
PRIVATE int atapiStop(ubyte WhichDrive)
{	ubyte Pkt[12]={
		ATAPI_CMD_STOP,0,0,0,0,0,0,0,0,0,0,0};
	struct drivecmd Cmd;

	Cmd.Blk=0;
	Cmd.Count=0;
	Cmd.Dev=WhichDrive;
	Cmd.Cmd=0;
	Cmd.Data=NULL;

	return(atapiCmd2(&Cmd, Pkt));
}

/****************************************************************************
        name:   atapiStatus                                                 
 ***************************************************************************/
PRIVATE int atapiStatus(ubyte WhichDrive)
{       ubyte Pkt[12]={0x00,0,0,0,0,0,0,0,0,0,0,0};
        struct drivecmd Cmd;

        Cmd.Blk=0;
        Cmd.Count=0;
        Cmd.Dev=WhichDrive;
        Cmd.Cmd=0;
        Cmd.Data=NULL;

        return(atapiCmd2(&Cmd, Pkt));
}
  
/****************************************************************************
        name:   atapiCapacity
        action: total cdrom capacity
        returns:whatever atapiCmd2() returns
****************************************************************************/
PRIVATE ulong atapiCapacity(ubyte WhichDrive)
{
    ubyte Pkt[12]={
            ATAPI_READ_CAPACITY,0,0,0,0,0,0,0,0,0,0,0};
    struct drivecmd Cmd;
    struct {
       ulong lba;
       ulong blocklen;
    }capbuf;
    int stat;
    Cmd.Blk=0;
    Cmd.Count=sizeof(capbuf);
    Cmd.Dev=WhichDrive;
    Cmd.Cmd=0;
    Cmd.Data=(ubyte *)&capbuf;

    stat=atapiCmd2(&Cmd, Pkt);
    if (stat==0) Capacity=ntohl(capbuf.lba);
    else Capacity=0x1fffff;
    return stat;
}

PRIVATE int atapiReadTime(ubyte WhichDrive,ubyte *Data)
{
   ubyte Pkt[12]={
           ATAPI_READ_SUB,0,0,0,0,0,0,0,0,0,0,0};
   struct drivecmd Cmd;
   int stat;

   Cmd.Blk=0;
   Cmd.Count=sizeof(struct atapi_subchnl);
   Cmd.Dev=WhichDrive;
   Cmd.Cmd=0;
   Cmd.Data=(ubyte *)Data;

   Pkt[1]=2; /* MSF Addressing */
   Pkt[2]=0x40; /* request subQ data */
   Pkt[3]=1;
   Pkt[7]=(Cmd.Count>>8);
   Pkt[8]=Cmd.Count;

   return(atapiCmd2(&Cmd, Pkt));
}

/*=========================================================================*
 *				c_handler 	  			   *	
 *=========================================================================*/
PRIVATE int c_handler(irq)
int irq; 	/* irq number */
{
  /* The interrupt handler, I never got an interrupt but its here just
   * in case...
   */

  /* Send interrupt message to cdrom task */
#if XXX
   interrupt(CDROM);
#endif
  interrupt(CDROM);

  return 1;
}

/*===========================================================================*
 *                              intr_wait                                    *
 *===========================================================================*/
PRIVATE int waitfor(mask, value)
int mask;
int value;
{
   struct milli_state ms;

   milli_start(&ms);
   do{
      if ((in_byte(Drive[2].IOAdr + ATA_REG_STAT) & mask) == value) return 1;
   }while (milli_elapsed(&ms) < TIMEOUT);
   return(0);
} 

/*===========================================================================*
 *				mcd_prepare				     *
 *===========================================================================*/
PRIVATE struct device *mcd_prepare(device)
int device;
{
  mcd_count = 0;

  /* Select partition. */
  if (device < DEV_PER_DRIVE) {			/* cd0, cd1, ... */
    mcd_dv = &mcd_part[device];
  } else
  if ((unsigned) (device -= MINOR_hd1a) < SUB_PER_DRIVE) { /* cd1a, cd1b, ... */
    mcd_dv = &mcd_subpart[device];
  } else {
    return NIL_DEV;
  }

  return mcd_dv;
}


/*===========================================================================*
 *				mcd_schedule				     *
 *===========================================================================*/
PRIVATE int mcd_schedule(proc_nr, iop)
int proc_nr;			/* process doing the request */
struct iorequest_s *iop;	/* pointer to read or write request */
{
/* Gather I/O requests on consecutive blocks so they may be read/written
 * in one controller command.  (There is enough time to compute the next
 * consecutive request while an unwanted block passes by.)
 */
   int r,opcode;
   unsigned long pos;
   unsigned nbytes;
   phys_bytes user_phys;

   nbytes = iop->io_nbytes;
    
   pos = iop->io_position;
    
   mcd_dv->dv_size=Capacity;

   user_phys = numap(proc_nr, (vir_bytes) iop->io_buf, nbytes);
   if (user_phys == 0) return (iop->io_nbytes = EINVAL);

   opcode = iop->io_request & ~OPTIONAL_IO;

   if (opcode!=DEV_READ) return EIO;

   if (pos >= mcd_dv->dv_size) return(OK);

   if (pos + nbytes > mcd_dv->dv_size) nbytes = mcd_dv->dv_size - pos;
   pos += mcd_dv->dv_base;

   mcd_tp = mcd_trans;

   mcd_tp->tr_iop = iop;
   mcd_tp->tr_pos = pos;
   mcd_tp->tr_count = nbytes;
   mcd_tp->tr_phys = user_phys;

   mcd_finish();

   /*mcd_tp++;
   mcd_count += nbytes;*/

   return(OK);
}


/*===========================================================================*
 *				mcd_finish				     *
 *===========================================================================*/
PRIVATE int mcd_finish()
{
/* Carry out the I/O requests gathered in mcd_trans[]. */
  struct trans *tp = mcd_trans;
  int r, errors;
  struct drivecmd cmd;
  uword IOAdr, Got;
  ubyte Phase;
  ulong Temp,pos;
  int i=0,n;
  unsigned count;
  struct drivecmd Cmd;
  ubyte Pkt[12]={ATAPI_CMD_READ10,0,0,0,0,0,0,0,0,0,0,0};

  /*if(mcd_count==0) return(OK);*/

  /*while(mcd_count>0)
  {*/
     pos = tp->tr_pos & ~MCD_BLOCK_MASK;

     count = (pos - tp->tr_pos) + mcd_count;
     count = (count + MCD_BLOCK_SIZE -1) & ~MCD_BLOCK_MASK;

     if (count>MCD_BLOCK_SIZE) count = MCD_BLOCK_SIZE;

     Cmd.Blk=16;
     Cmd.Count=2048;

     Pkt[1]=0;
     Pkt[2]=Cmd.Blk >> 24;
     Pkt[3]=Cmd.Blk >> 16;
     Pkt[4]=Cmd.Blk >> 8;
     Pkt[5]=Cmd.Blk;
     Pkt[6]=Cmd.Count >> 16;
     Pkt[7]=Cmd.Count >> 8;
     Pkt[8]=Cmd.Count;
     Pkt[9]=Pkt[10]=Pkt[11]=0;

     IOAdr=Drive[2].IOAdr;

     if(ataSelect(IOAdr, Drive[2].DrvSel)) {    return(-1); } 
     milli_delay(400);
      
	out_byte(IOAdr + ATA_REG_FEAT, 0);	/* irrelevant? */
	out_byte(IOAdr + ATA_REG_REASON, 0);/* irrelevant? */
	out_byte(IOAdr + ATA_REG_SECT, 0);	/* irrelevant? */
	out_byte(IOAdr + ATA_REG_LOCNT, 32768ul);
	out_byte(IOAdr + ATA_REG_HICNT, 32768ul >> 8);
	out_byte(IOAdr + ATA_REG_CMD, ATA_CMD_PKT);

	for(Temp=500; Temp; Temp--)	/* XXX - why 500? */
	{	if((in_byte(IOAdr + ATA_REG_STAT) & 0x88) == 0x08) break;
		milli_delay(1); }	/* this _must_ be polled, I guess */
	if(Temp == 0)
	{   return(-3); }
	milli_delay(400);

	outsw(IOAdr + ATA_REG_DATA, (uword *)Pkt, 6);
	milli_delay(400);

	/* For each sector, wait for an interrupt and fetch the data (read),
	 * or supply data to the controller and wait for an interrupt (write).
	 */

	/*while(cd_count>0 && count>0)
	{*/
	    waitfor(0x80,0);
	     
	    Phase=in_byte(IOAdr + ATA_REG_STAT) & 0x08; 
	    Phase |= (in_byte(IOAdr + ATA_REG_REASON) & 3);
	    milli_delay(200);
	    if(Phase == ATAPI_PH_DONE) {if(Cmd.Count) { return(-5);} return(0);}
	    else if(Phase == ATAPI_PH_ABORT) { return(-6); }
	    else if(Phase != ATAPI_PH_DATAIN) {	printf("atapiCmd2: bad phase %u\n", Phase); return(-7); }

	    Got=in_byte(IOAdr + ATA_REG_HICNT);
	    Got=(Got << 8) | in_byte(IOAdr + ATA_REG_LOCNT);
	    Temp=MIN(Cmd.Count,Got);

	    /*n=tp->tr_count;
	    if(n>count) n=count;*/
	    milli_delay(200);

	    port_read(IOAdr + ATA_REG_DATA, tp->tr_phys, 2048);

	    milli_delay(200);

	    /*for(Got-=Temp;Got>1;Got-=2) in_word(IOAdr + ATA_REG_DATA);
	    if(Got) in_byte(IOAdr + ATA_REG_DATA);*/

	    tp->tr_phys += 2048; 
	    tp->tr_iop->io_nbytes -= 2048;
	    if((tp->tr_count -= Temp) ==0) tp++;
  return(OK);
}


/*============================================================================*
 *				mcd_geometry				      *
 *============================================================================*/
PRIVATE void mcd_geometry(entry)
struct partition *entry;		
{
/* The geometry of a cdrom doesn't look like the geometry of a regular disk,
 * so we invent a geometry to keep external programs happy.
 */ 
  entry->cylinders = (mcd_part[0].dv_size >> SECTOR_SHIFT) / (64 * 32);
  entry->heads = 64;
  entry->sectors = 32;
}

#endif /* ENABLE_ATAPI_CDROM */

