/* This file contains the driver for an Atapi cdrom controller.
 *
 * The file contains one entry point:
 *
 *   mcd_task:		main entry when system is brought up
 *
 *  Mar 12 1999			Author: Alexis LEDOUX 
 */


/*////////////////////////////////////////////////////////////////////////////
				                ATAPI STUFF
////////////////////////////////////////////////////////////////////////////*/
typedef unsigned char	ubyte;
typedef unsigned short	uword;
typedef unsigned long	ulong;

#define put_unaligned(val,ptr) ((void)((*ptr)=val))
#define get_unaligned(ptr) (*(ptr))

#define CDPLAY		0
#define CDSTOP		1
#define CDPAUSE		2
#define CDNEXT		3
#define CDPREVIOUS	4
#define CDEJECT		5
#define YES		1
#define NO		0
#define CDSIZE		100

/* Delays from Linux ide.h */
#define	WAIT_READY	30	/* RDY asserted, use 5000 for notebook/APM */
#define	WAIT_PID	3	/* ATAPI device responds to 'identify' */
#define	WAIT_CMD	10000	/* IRQ occurs in response to command */
#define DRV_CMD_RD	1
#define DRV_CMD_WR	2

/* ATAPI sector size */
#define	ATAPI_LG_SECTSIZE	11	/* 2K CD-ROM sectors */
#define	ATAPI_SECTSIZE		(1 << (ATAPI_LG_SECTSIZE))

/* ATA drive command bytes */
#define	ATA_CMD_PKT	0xA0		/* ATAPI packet cmd */
#define	ATA_CMD_PID	0xA1		/* ATAPI identify */

/* ATAPI drive register file */
#define	ATA_REG_DATA	0		/* data (16-bit) */
#define	ATA_REG_FEAT	1		/* write: feature reg */
#define	ATA_REG_ERR	ATA_REG_FEAT	/* read: error */
#define	ATA_REG_CNT	2		/* ATA: sector count */
#define	ATA_REG_REASON	ATA_REG_CNT	/* ATAPI: interrupt reason */
#define	ATA_REG_SECT	3		/* sector */
#define	ATA_REG_LOCYL	4		/* ATA: LSB of cylinder */
#define	ATA_REG_LOCNT	ATA_REG_LOCYL	/* ATAPI: LSB of transfer count */
#define	ATA_REG_HICYL	5		/* ATA: MSB of cylinder */
#define	ATA_REG_HICNT	ATA_REG_HICYL	/* ATAPI: MSB of transfer count */
#define	ATA_REG_DRVHD	6		/* drive select; head */
#define	ATA_REG_CMD	7		/* write: drive command */
#define	ATA_REG_STAT	7		/* read: status and error flags */
#define	ATA_REG_SLCT	0x206		/* write: device control */
#define	ATA_REG_ALTST	0x206		/* read: alternate status/error */

/* ATAPI packet command bytes */
#define	ATAPI_CMD_START_STOP	0x1B	/* eject/load */
#define ATAPI_CMD_LOCK_UNLOCK	0x1E	/* Lock / Unlock the door */
#define	ATAPI_CMD_READ10	0x28	/* read data sector(s) */
#define ATAPI_CMD_READ12	0xA8	/* read data sector(s) */
#define	ATAPI_CMD_READTOC	0x43	/* read audio table-of-contents */
#define	ATAPI_CMD_PLAY		0x47	/* play audio */
#define	ATAPI_CMD_PAUSE		0x4B	/* pause/continue audio */
#define ATAPI_CMD_STOP		0x4E    /* Stop the play */
#define ATAPI_CMD_SPEED		0xBB    /* Speed of read and write */
#define ATAPI_READ_CAPACITY	0x25	/* Read total Capacity of CDROM */
#define ATAPI_READ_SUB		0x42	/* Read sub channel */
#define ATAPI_MODE_SENSE_10	0x5A
#define ATAPI_MODE_SELECT_10	0x55
#define ATAPI_READCD		0xBE

#define PAGE_AUDIO		0x0E

/* ATAPI data/command transfer 'phases' */
#define	ATAPI_PH_ABORT		0	/* other possible phases (1, 2, */
#define	ATAPI_PH_DONE		3	/* and 11) are invalid */
#define	ATAPI_PH_DATAOUT	8
#define	ATAPI_PH_CMDOUT		9
#define	ATAPI_PH_DATAIN		10

#define ISO9660_ID	"CD001"
#define HIGH_SIERRA_ID	"CDROM"
#define ATAPI_BLOCK_SIZE	2048
#define BLOCK_SHIFT	11

#define NULL_DIR	(struct dir_entry *) 0
#define MAX_NAME_LENGTH	255
#define MAX_PATH_LENGTH	1024

#define NR_ELT_PER_COL	19
#define NR_OF_COLS	2
#define NR_OF_CHARS	13	

/* This macro always returns a lower case character */
#define LOWER_CASE(CHR) (CHR >= 'A' && CHR <= 'Z' ? CHR | 0x20 : CHR)

/* Macro's for determining . , .. and normal directory entries */
#define IS_DOT(PTR) (PTR->name_length[0] == 1 && PTR->name[0] == 0 ? 1 : 0)
#define IS_DOT_DOT(PTR) (PTR->name_length[0] == 1 && PTR->name[0] == 1 ? 1 : 0)
#define IS_DIR(PTR) (PTR->flags[-High_Sierra] & 2 ? 1 : 0)

/* Constant for the state */
#define OLD		1
#define NEW		0
#define YES		1
#define NO		0
#define DIRECTORY	1
#define NOTDIRECTORY	2
#define CDAUDIO		3	

/* generalized drive info structure */
struct drive
{	ubyte DrvSel;		/* ATAPI only (LUN for SCSI?) */
	uword IOAdr;
};

/* ATAPI command structure */
struct drivecmd
{	ulong Blk;	/* in SECTORS */
	ulong Count;	/* in BYTES */
	ubyte Dev, Cmd, *Data;
};

struct ataid     /* 'identify' structure, as per ANSI ATA2 rev.2f spec */
{	uword Config, PhysCyls, Res2, PhysHeads, UnfBytesPerTrack;
	uword UnfBytesPerSect, PhysSects, Vendor0, Vendor1, Vendor2;
	ubyte SerialNum[20];
	uword BufType, BufSize, ECCBytes;
	ubyte FirmwareRev[8], Model[40], MaxMult, Vendor3;
	uword DwordIO;
	ubyte Vendor4, Capability;
	uword Res50;
	ubyte Vendor5, PIOMode, Vendor6, DMAMode;
	uword LogValid, LogCyls, LogHeads, LogSects;
	ulong TotalSects;
	ubyte MultSect, MultSectValid;
	ulong LBASects;
	uword DMAInfoSingle, DMAInfoMult, EIDEPIOModes;
	uword EIDEDMAMin, EIDEDMATime, EIDEPIO, EIDEPIOIORdy;
	uword Res[187];
};

struct atapimsf		/* 4 bytes */
{	ubyte Res0, Min, Sec, Frame; };

struct atapitocheader	/* 4 bytes */
{	uword TocLen;
	ubyte FirstTrk;
	ubyte LastTrk; };

struct atapitocentry	/* 8 bytes */
{	ubyte Res0;
	/*ubyte Ctrl;*/
	ubyte Adr;     /*0h = information not supplied
		       //1h = current position number
		       //2h = encodes media catalogue number
		       //3h encdes ISRC
		       //else reserved */
	ubyte Trk;     /*pro 00h audio without pre-emphasis
		       //pro 01h audio with pre-emphasis
		       //per 02h audio without
		       //per 03h audio with
		       //pro 04h digital data
		       //pro 05h reserved
		       //per 06h digital data
		       //per 07h reserved
		       //else broadcast use */
	ubyte Res1;
	union
	{	ulong Block;
		struct atapimsf Time; } Where; };

struct atapitoc
{	struct atapitocheader Hdr;
	struct atapitocentry Ent[100]; 
};

struct volctrl
{
   ubyte channel0;
   ubyte channel1;
   ubyte channel2;
   ubyte channel3;
};

struct data_read
{ 
   long Offset;
   ubyte Data[2048];
};

struct atapi_subchnl
{
  ubyte reserved;
  ubyte audio_status;
  uword length;
  ubyte format;

  ubyte adr_and_ctrl;
  ubyte trk;
  ubyte ind;
  union {
     struct atapimsf msf;
     int lba;
  }absaddr;
  union {
     struct atapimsf msf;
     int lba;
  }reladdr;
};

struct cd {
   int start;
   int end;
   int Numtrack;
   char title[40];
   char track[55][60];
};

/* Fields in a ISO9660 volume descriptor */
struct iso9660_descriptor
{
  char type[1];		
  char id[5];			
  char version[1];		
  char reserved1[1];		
  char system_id[32];		
  char volume_id[32];		
  char reserved2[8];		
  char volume_size[8];	
  char reserved3[32];		
  char volume_set_size[4];	
  char volume_seq_nr[4];
  char block_size[4];	
  char path_table_size[8];	
  char type_l_path_table[4];	
  char opt_type_l_path_table[4];	
  char type_m_path_table[4];	
  char opt_type_m_path_table[4];	
  char root_dir_entry[34];
  char vol_set_id[128];	
  char publ_id[128];
  char prep_id[128];	
  char appl_id[128];	
  char copyright_file_id[37];	
  char abstract_file_id[37];	
  char bibl_file_id[37];
  char creation_date[17];	
  char mod_date[17];	
  char exp_date[17];	
  char eff_date[17];	
  char file_struc_version[1];
  char reserved4[1];		
  char appl_data[512];	
  char reserved5[653];			
};


/* Fields in a High Sierra volume descriptor */
struct high_sierra_descriptor
{
  char reserved1[8];
  char type[1];	
  char id[5];		
  char version[1];	
  char reserved2[1];	
  char system_id[32];
  char volume_id[32];
  char reserved3[8];	
  char volume_size[8];	
  char reserved4[32];
  char vol_set_size[4];
  char volume_seq_nr[4];	
  char block_size[4];	
  char path_table_size[8];
  char type_l_path_table[4];
  char reserved5[28];		
  char root_dir_entry[34];
};


/* Fields in a directory entry */
struct dir_entry 
{
  char length[1];	
  char ext_attr_length[1];
  char first_block[8];
  char size[8];	
  char date[7];
  char flags[1];	
  char file_unit_size[1];
  char interleave[1];
  char volume_seq_nr[4];	
  char name_length[1];
  char name[1];
};
